# Copyright European Space Agency, 2013

from __future__ import division

import numpy as np
import numpy.ma as ma
from astropy.coordinates import Angle
import astropy.units as u

import unittest
from nose.plugins.attrib import attr

from auromat.resample import resample, resampleMLatMLT
from auromat.mapping.mapping import GenericMapping, checkPlateCarree
import os
from auromat.mapping.spacecraft import getMapping
from auromat.coordinates.transform import mltToSmLon, rotatePole
import datetime
from numpy.ma.testutils import assert_array_approx_equal

def _testCoords(offset):
    n = 10
    sp, step = np.linspace(offset,offset+10, num=n, retstep=True)
    coord = np.tile(sp, n).reshape(n,n).astype(np.float32)

    r = n*0.4
    y,x = np.ogrid[-r: r+1, -r: r+1]
    mask = x**2+y**2 <= r**2

    coord[mask] = -coord[mask]
    coord[coord>0] = np.nan
    coord[mask] = -coord[mask]
    
    coordCenter = coord[:-1,:-1] + step/2
    
    return coord, coordCenter
    
def testCoords():
    # no discontinuity or pole
    lats, latsCenter = _testCoords(70)
    lats = lats.T
    latsCenter = latsCenter.T
    
    lons, lonsCenter = _testCoords(160)
    
    return lats, lons, latsCenter, lonsCenter
    
def testCoordsDiscontinuity():
    lats, latsCenter = _testCoords(70)
    lats = lats.T
    latsCenter = latsCenter.T
    
    lons, lonsCenter = _testCoords(160)
    
    lons = Angle((lons + 15) * u.deg).wrap_at(180 * u.deg).degree
    lonsCenter = Angle((lonsCenter + 15) * u.deg).wrap_at(180 * u.deg).degree
    return lats, lons, latsCenter, lonsCenter

def testCoordsPole():
    lons, lonsCenter = _testCoords(1)
    lons = Angle((lons - 5) * u.deg).wrap_at(180 * u.deg).degree
    lonsCenter = Angle((lonsCenter - 5) * u.deg).wrap_at(180 * u.deg).degree
    lats = lons.T
    latsCenter = lonsCenter.T
    latsRot, lonsRot = rotatePole(np.deg2rad(lats.flat), np.deg2rad(lons.flat), 0, angle=90, axis=[0,1,0])
    latsCenterRot, lonsCenterRot = rotatePole(np.deg2rad(latsCenter.flat), np.deg2rad(lonsCenter.flat), 0, angle=90, axis=[0,1,0])
    return np.rad2deg(latsRot.reshape(lats.shape)), np.rad2deg(lonsRot.reshape(lons.shape)),\
           np.rad2deg(latsCenterRot.reshape(latsCenter.shape)), np.rad2deg(lonsCenterRot.reshape(lonsCenter.shape))

class Test(unittest.TestCase):
    def test(self):
        np.set_printoptions(precision=1)
        lats, lons, latsCenter, lonsCenter = testCoordsDiscontinuity()

        altitude = 110
        rgb = ma.masked_array((np.random.rand(lats.shape[0]-1,lats.shape[1]-1,3) * 255).astype(np.uint8))
        elevation = np.zeros((rgb.shape[0], rgb.shape[1]))
                
        mapping = GenericMapping(lats, lons, latsCenter, lonsCenter, elevation, altitude, rgb, 
                                 cameraPosGCRS=np.array([0,0,0]), photoTime=datetime.datetime.now(), identifier=None)
        m = resample(mapping, pxPerDeg=1, method='mean')
        m.checkPlateCarree()
        
        mapping = resampleMLatMLT(mapping, arcsecPerPx=100, method='nearest')
        assert not mapping.isPlateCarree
        mlat, mlt = mapping.mLatMlt
        smlon = mltToSmLon(mlt)
        checkPlateCarree(mlat, smlon)
    
    @attr('slow')          
    def testReal(self):
        m1 = _getMapping()
        m2 = resample(m1, pxPerDeg=15, method='mean')
        m2.checkPlateCarree()
        m3 = resample(m2, arcsecPerPx=100, method='nearest')
        m3.checkPlateCarree()
        
        assert_array_approx_equal(_bbToArray(m2.boundingBox), _bbToArray(m1.boundingBox), 1)
        assert_array_approx_equal(_bbToArray(m3.boundingBox), _bbToArray(m1.boundingBox), 1)
    
    @attr('slow')    
    def testPoleBug(self):
        m1 = _getMapping()
        m2 = resample(m1, arcsecPerPx=100, method='mean')
        assert_array_approx_equal(_bbToArray(m2.boundingBox), _bbToArray(m1.boundingBox), 1)
        
def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

def _getMapping():
    imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
    wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def _bbToArray(bb):
    return [bb.latNorth, bb.latSouth, bb.lonWest, bb.lonEast]



# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

import logging
import os
from six.moves.urllib.parse import urlencode
from six.moves import range
from datetime import datetime, timedelta

import ephem

from auromat.coordinates.ephem import EphemerisCalculator
from auromat.util.os import makedirs
import auromat.util.url

baseUrl = 'https://www.space-track.org/'
authUrl = baseUrl + 'ajaxauth/login'
queryPrefix = 'basicspacedata/query/'

class Spacetrack:
    """
    Downloads TLE data from http://space-track.org and stores them as files
    on disk (one file per NORAD ID).
    
    Use the :mod:`auromat.coordinates.ephem` module for calculating
    coordinates based on the TLE files.
    """
    
    def __init__(self, user, password, minUpdateInterval=timedelta(days=30)):
        """
        
        :param str user: space-track.org user name
        :param str password: space-track.org password
        :param datetime.timedelta minUpdateInterval:
            The minimum amount of time after which an update of a cached
            TLE file is performed when using one of the updateTLEs* methods.
            The time is relative to the file modification dates.
        """
        self.user = user
        self.password = password
        self.minUpdateInterval = minUpdateInterval
    
    def updateTLEsFor(self, noradId, tlePath, observationDate):
        """
        Updates the TLEs such that the given observation date is covered.
        If no suitable TLEs are available (newest TLE Epoch < observationDate),
        an exception is raised.
        
        Note that TLEs are only appended, which means that TLE files
        which were produced outside this class and which only contain
        a certain time period will not be prepended with older TLEs.
        
        :param str|int noradId:
        :param str tlePath:
        :param datetime.datetime observationDate:
        :raise: DownloadError: on any network error
        :raise: ValueError: if the downloaded TLEs could not be correctly read
        :raise: ValueError: if no TLEs were found for the given date
        """
        noradId = int(noradId)
        
        if not os.path.exists(tlePath):
            self.updateTLEs(noradId, tlePath)
            newTleFile = True
        else:
            newTleFile = False

        observationTimeOutsideTLERange = False
        
        ephemCalculator = EphemerisCalculator(tlePath)
        if ephemCalculator.lastEpoch < observationDate:
            if not newTleFile:
                newTlesAdded = self.updateTLEs(noradId, tlePath)
                if newTlesAdded:
                    ephemCalculator = EphemerisCalculator(tlePath)
                    if ephemCalculator.lastEpoch < observationDate:
                        observationTimeOutsideTLERange = True
                else:
                    observationTimeOutsideTLERange = True
            else:
                observationTimeOutsideTLERange = True
        
        elif ephemCalculator.firstEpoch > observationDate:
            # Here we could check whether there are older TLEs available than already stored.
            # We don't do it as this won't be the case if this class was used for creating the
            # existing TLE file. (assuming that spacetrack doesn't release older TLEs after newer
            # ones are released)
            observationTimeOutsideTLERange = True
        
        if observationTimeOutsideTLERange:
            raise ValueError('No TLEs (NORAD: ' + str(noradId) + ') available >= ' + str(observationDate))
    
    def updateTLEs(self, noradId, tlePath):
        """
        Updates the TLEs to the latest available data.
        
        :param str|int noradId:
        :param str tlePath:
        :return: True, if new TLEs were added, False otherwise
        :raise: DownloadError: on any network error
        :raise: ValueError: if the downloaded TLEs could not be correctly read
        """
        if os.path.exists(tlePath):
            mtime = datetime.fromtimestamp(os.path.getmtime(tlePath))
            if datetime.now() - mtime < self.minUpdateInterval:
                return False
            
            # read latest available epoch
            with open(tlePath, 'r') as t:
                tles = t.readlines()
            lastTle = ephem.readtle('foo', tles[-2], tles[-1])
            year,month,day,h,m,s = lastTle._epoch.tuple()
            date = datetime(year,month,day,h,m,int(round(s))).strftime('%Y-%m-%d %H:%M:%S')
        else:
            date = '0000-00-00'
            tles = []

        query = 'class/tle/NORAD_CAT_ID/%s/EPOCH/>%s/orderby/EPOCH asc/format/tle' % (noradId,date)
        response = self.query(query)
        newTles = response.splitlines()
                
        if len(newTles) == 0:
            return False
        
        if len(newTles) % 2 != 0:
            raise ValueError('The number of returned TLE lines from space-track.org is not a multiple of 2')
        
        # filter out TLEs where the checksum is missing
        # e.g. within the ISS TLEs sporadically between 2001-2004
        # Note that appending a 0 as checksum isn't enough to satisfy pyephem,
        # but it would be enough for the sgp4 library.
        # TODO recalculate checksum
        newTles = [line for line in newTles if len(line) == 69]
                
        tleCount = len(newTles) // 2
        for i in range(tleCount):
            try:
                ephem.readtle('foo', newTles[i*2], newTles[i*2 + 1])
            except Exception as e:
                raise ValueError("The following TLE couldn't be read: [" + 
                                  newTles[i*2] + ', ' + newTles[i*2+1] + '] (reason: ' + repr(e) + ')')
        
        makedirs(os.path.dirname(tlePath))
        with open(tlePath, 'a') as t:
            t.write('\n'.join(newTles))
            t.write('\n')
        
        return True
        
    def query(self, query):
        """
        Query spacetrack and return the result as string. 
        
        :raise DownloadError: on any network error
        """
        queryUrl = baseUrl + queryPrefix + query
        logging.info('Querying ' + queryUrl)
        
        data = urlencode({'identity': self.user, 'password': self.password, 'query': queryUrl})
        res = auromat.util.url.downloadResource(authUrl, lambda r: r.read(), data=data)
        return res

if __name__ == '__main__':
    import sys
    
    assert len(sys.argv) == 5, 'Syntax: spacetrack.py user pass noradid tlepath' 
        
    st = Spacetrack(user=sys.argv[1], password=sys.argv[2])
    if st.updateTLEs(int(sys.argv[3]), sys.argv[4]):
        print('new TLEs were added')
    else:
        print('no updates available or download/read error')
        


"""
This package allows to georeference images using the camera position and
the starfield within the image. It should be generic enough to work with
any kind of images but was developed with ISS aurora images in mind.
It tries to automatically detect where the starfield in the image is
(see :mod:`auromat.solving.masking` modules) and then
runs astrometry.net to extract the stars and find the pointing and scale of the
image (see :mod:`auromat.solving.solving` module).

For images taken from spacecrafts (like the ISS) it contains two convenience
modules: :mod:`~auromat.solving.spacecraft` and :mod:`~auromat.solving.eol`.
The :mod:`~auromat.solving.spacecraft` module
determines the spacecraft position (=camera position) at the time
the image was taken using TLE data from space-track.org, given the NORAD ID of the spacecraft.
The :mod:`~auromat.solving.eol` module allows to easily download images
from `NASA's Earth Observation website <http://eol.jsc.nasa.gov>`_ in JPEG and RAW format.

The resulting calibration is a FITS WCS header from which the geodetic
pixel coordinates can be calculated from. The counterpart modules in the
mapping package are :mod:`auromat.mapping.astrometry` and :mod:`auromat.mapping.spacecraft`.
The :mod:`auromat.mapping.iss` module relies on those two modules as well and
is tailored towards the ESA ISS archive containing such images and calibrations.
"""


# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import unittest
from nose.plugins.attrib import attr

from auromat.utils import _outline_opencv, _outline_skimage, polygonArea,\
    polygonCentroid, _polygonCentroidOpenCV
from numpy.testing.utils import assert_almost_equal
from auromat.mapping.spacecraft import getMapping
import os.path

try:
    import cv2 as cv
except ImportError:
    cv = None

def _testIm(n=10):
    coord = np.ones((n,n))

    r = n*0.4
    y,x = np.ogrid[-r: r+1, -r: r+1]
    mask = x**2+y**2 <= r**2

    coord[mask] = -coord[mask]
    coord[coord>0] = np.nan
    # make it non-symmetric
    coord[4,0] = np.nan
    
    return ~np.isnan(coord)

if __name__ == '__main__':
    np.set_printoptions(precision=1)
    
    
#    def testImage():
#        n = 10
#        coord = np.ones((n,n))
#    
#        r = n*0.4
#        y,x = np.ogrid[-r: r+1, -r: r+1]
#        mask = x**2+y**2 <= r**2
#    
#        coord[mask] = -coord[mask]
#        coord[coord>0] = np.nan
#        # make it non-symmetric
#        coord[4,0] = np.nan
#        
#        r = ~np.isnan(coord)
#        return np.hstack((r,r))
#    
#    import skimage.measure
#    contours = skimage.measure.find_contours(testImage(), 0.99)
#    print contours
#    raise
    
    
    
    im = _testIm(3000)
    print(im)
    
    outl = _outline_opencv(im)
    outlsk = _outline_skimage(im)
    
    print(len(outl) == len(outlsk))
    print('opencv:', outl)
    print('skimage:', outlsk)
    
    
    # TODO check that both outlines are the same modulo starting point
    
    for title, p in [('skimage', outlsk), ('opencv', outl)]:
        fig, ax = plt.subplots()
        ax.imshow(im, interpolation='nearest', cmap=plt.cm.gray)
        ax.plot(p[:, 0], p[:, 1], linewidth=2)
        ax.set_title(title)
        ax.axis('image')
        ax.set_xticks([])
        ax.set_yticks([])
    
    
    # TODO test that biggest contour is returned
    print('TWO CONTOURS')
    im2 = np.hstack((im, im))
    print(im2)
    
    outl = _outline_opencv(im2)
    outlsk = _outline_skimage(im2)
    print(len(outl) == len(outlsk))
    print('opencv:', outl)
    print('skimage:', outlsk)
    
    # TODO check that both outlines are the same modulo starting point
    for title, p in [('skimage-2', outlsk), ('opencv-2', outl)]:
        fig, ax = plt.subplots()
        ax.imshow(im2, interpolation='nearest', cmap=plt.cm.gray)
#        ax.plot(p[:, 0], p[:, 1], linewidth=2)
#        ax.set_title(title)
        ax.axis('image')
        ax.set_xticks([])
        ax.set_yticks([])
    plt.show()
    
class Test(unittest.TestCase):
    def testPolygonArea(self):
        poly = [[4, 8],
               [3, 7],
               [2, 7],
               [1, 6],
               [1, 5],
               [1, 4],
               [1, 3],
               [1, 2],
               [2, 1],
               [3, 1],
               [4, 0],
               [5, 1],
               [6, 1],
               [7, 2],
               [7, 3],
               [8, 4],
               [7, 5],
               [7, 6],
               [6, 7],
               [5, 7]]
    
        assert polygonArea(poly) == 37.0
        
    def testPolygonCentroid(self):
        
        poly = [
            (30,50),
            (200,10),
            (250,50),
            (350,100),
            (200,180),
            (100,140),
            (10,200)
            ]
        
        assert_almost_equal(polygonCentroid(poly), (159.2903828197946, 98.88888888888))
        
        if cv:
            assert_almost_equal(polygonCentroid(poly), _polygonCentroidOpenCV(poly))
    
    @attr('slow')
    def testMappingCentroid(self):
        m = _getMapping()
        # for many polygon points like here, the two solutions (opencv using moments and
        # plain python) differ slightly after the 7th decimal
        # for our purposes this should still be accurate enough
        assert_almost_equal(m.centroid, 
                            [55.00295889563608, -99.21825084682715], # using cv.moments 
                            decimal=6)
        
def _getMapping():
    imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
    wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)
    


# Copyright European Space Agency, 2013

"""
A collection of utility functions which should eventually become part
of the :mod:`auromat.util` package.
"""

from __future__ import division, print_function, absolute_import

import logging
import bisect
from itertools import groupby

import numpy as np
from numpy.core.umath_tests import inner1d # vector-wise dot product
import itertools

from scipy.spatial import Delaunay

import skimage.measure

import matplotlib.path
try:
    import matplotlib.nxutils
except:
    pass
            
def vectorLengths(vectors):
    """ `np.linalg.norm(vectors, axis=1)` for numpy < 1.8. """
    vectors = np.asarray(vectors)
    return np.sqrt((vectors*vectors).sum(axis=1))

def unitVectors(vectors):
    """ Return the unit vectors of an array of vectors. """
    vectors = np.asarray(vectors)
    return vectors / vectorLengths(vectors)[...,None]

def angleBetween(v1, v2):
    """ Return the angles in radians between two unit vector arrays.
        Angles are in [0,pi]. """
    # When vectors are equal their dot product is 1, but
    # due to rounding it can be slightly above 1 which would then
    # result in arccos returning NaN. Therefore we clip the values to [-1,1]
    dot = np.clip(inner1d(v1, v2), -1, 1)
    angle = np.arccos(dot)
    return angle

def signedAngleBetween(v1, v2):
    """ 
    Return the angles in radians between two 2D vector arrays.
    Angles are in [-pi,pi].
    
    :see: http://stackoverflow.com/a/2150475
    """
    angle = np.arctan2(v1[:,0]*v2[:,1] - v1[:,1]*v2[:,0], v1[:,0]*v2[:,0] + v1[:,1]*v2[:,1])
    return angle

def pointsInsidePolygon(points, polygon):
    """
    Return for each point if it lies inside the given polygon.
    
    :param points: shape (n,2)
    :param polygon: unclosed, shape (n,2)
    :rtype: boolean ndarray of shape (n,)
    """
    path = matplotlib.path.Path(polygon)
    try:
        # only in mpl >= 1.2.0        
        isInside = path.contains_points(points)
    except AttributeError:
        # we are on an older version, try nxutils instead
        # nxutils is deprecated in mpl >= 1.2.0
        isInside = matplotlib.nxutils.points_inside_poly(points, polygon)
    return isInside

def _outline_opencv(im):
    import cv2 as cv
    # "Also, the function does not take into account 1-pixel border of the image
    # (it's filled with 0's and used for neighbor analysis in the algorithm),
    # therefore the contours touching the image border will be clipped."
    # (findContours docs)
    # -> as we want the borders, we pad it with 1px on each side
    imCv = np.zeros((im.shape[0]+2, im.shape[1]+2), dtype=np.uint8)
    imCv[1:-1,1:-1] = im
    # cv.CHAIN_APPROX_SIMPLE *must not* be used here, as it would turn concave parts
    # into convex parts!
    contours,_ = cv.findContours(imCv, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)
    
    if len(contours) > 1:
        area = list(map(cv.contourArea, contours))
        contour = contours[np.argmax(area)]
        logging.warn('The binary data contains multiple contours! Returning the biggest one.')
    else:
        contour = contours[0]
    return np.asarray(contour).reshape(-1,2) - 1

def _outline_skimage(im):
    imBorder = np.zeros((im.shape[0]+2, im.shape[1]+2), dtype=bool)
    imBorder[1:-1,1:-1] = im
    # by using a level near 1 (=True) we get coordinates that can
    # be directly used as array indices (after rounding)
    level = 0.99
    contours = skimage.measure.find_contours(imBorder, level)
    
    def _fixContour(contour):
        # contrary to the API docs, the returned coordinates are (y,x)
        # instead of (x,y)
        # see https://github.com/scikit-image/scikit-image/issues/1140
        contour = np.fliplr(contour)
        
        # subtract padding and round to int
        contour = np.round(contour-1).astype(int)
        
        # remove consecutive duplicate values
        # TODO why are these here in the first place?
        contour = withoutConsecutiveDuplicates(contour)
    
        # un-close polygon
        contour = contour[:-1]
               
        return contour
        
    if len(contours) > 1:
        contours = list(map(_fixContour, contours))
        # degenerate contours may result from an initial contour such as
        # [[ 186.    320.01]
        #  [ 185.99  320.  ]
        #  [ 186.    319.99]
        #  [ 186.01  320.  ]
        #  [ 186.    320.01]]
        # after _fixContour() the contour is empty
        contours = list(filter(lambda c: len(c) > 2, contours))
        area = list(map(polygonArea, contours))
        contour = contours[np.argmax(area)]
        logging.warn('The binary data contains multiple contours! Returning the biggest one.')
    else:
        contour = _fixContour(contours[0])
    
    return contour

def outline(im):
    """ 
    Finds the outline of a binary image, assuming that the inner structure
    is filled with True's. The returned points are in clockwise order and can
    be used as a polygon.
    This works for concave forms as well.
    
    :param im: shape (h,w)
    :rtype: ndarray of shape (n,2) in x,y order
    """
    return _outline_skimage(im)

def polygonArea(poly, signed=False):
    """
    Return area of an unclosed polygon.
    
    :see: http://www.mathopenref.com/coordpolygonarea.html
    :param poly: (n,2)-array
    """
    # There is no equivalent function in skimage yet.
    # See https://groups.google.com/d/topic/scikit-image/--pXW-fPbGg/discussion
    
    # we need a plain list for the following operations
    if isinstance(poly, np.ndarray):
        poly = poly.tolist()
    segments = zip(poly, poly[1:] + [poly[0]])
    area = 0.5 * sum(x0*y1 - x1*y0
                     for ((x0, y0), (x1, y1)) in segments)
    if not signed:
        area = abs(area)
    return area

def polygonCentroid(poly):
    """
    Return centroid point of an unclosed polygon.
    
    :param poly: (n,2)-array
    :rtype: tuple (x,y)
    """
    # see https://stackoverflow.com/a/14128955
    #Copyright (c) 2013, Stack Overflow user "unutbu"
    #All rights reserved.
    #
    #Redistribution and use in source and binary forms, with or without modification,
    #are permitted provided that the following conditions are met:
    #
    #1. Redistributions of source code must retain the above copyright notice,
    #   this list of conditions and the following disclaimer.
    #
    #2. Redistributions in binary form must reproduce the above copyright notice,
    #   this list of conditions and the following disclaimer in the documentation 
    #   and/or other materials provided with the distribution.
    #
    #3. Neither the name of the copyright holder nor the names of its contributors 
    #   may be used to endorse or promote products derived from this software
    #   without specific prior written permission.
    #
    #THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
    #AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
    #IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
    #ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
    #LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    #DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
    #SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
    #CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
    #OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
    #OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    area = polygonArea(poly)
    
    result_x = 0
    result_y = 0
    N = len(poly)
    points = itertools.cycle(poly)
    x1, y1 = next(points)
    for _ in range(N):
        x0, y0 = x1, y1
        x1, y1 = next(points)
        cross = (x0 * y1) - (x1 * y0)
        result_x += (x0 + x1) * cross
        result_y += (y0 + y1) * cross
    result_x /= (area * 6.0)
    result_y /= (area * 6.0)
    return (result_x, result_y)

def _polygonCentroidOpenCV(poly):
    import cv2
    poly = np.require(poly, np.float32, 'C')
    m = cv2.moments(poly.reshape(-1,1,2))
    x = m['m10']/m['m00']
    y = m['m01']/m['m00']
    return x, y

def withoutConsecutiveDuplicates(arr):
    """
    Return a copy of the input array where consecutive duplicates
    (on the first dimension) are removed.
    
    :param arr: list or ndarray
    :rtype: ndarray
    """
    if isinstance(arr, np.ndarray):
        arr = arr.tolist()
    return np.asarray([p[0] for p in groupby(arr)])

def convexHull(points):
    """
    Return the convex hull spanning the given points.
    
    :param points: array of shape (n,2)
    :rtype: array of shape (m,2)
    """
    points = np.asarray(points)
    assert points.ndim == 2 and points.shape[1] == 2

    # TODO using scipy.spatial.ConvexHull lead to inconsistent results
    #      -> possible bug? for now, just use Delaunay
#    if ConvexHull:
#        # scipy.spatial.ConvexHull was introduced in 0.12
#        hull = ConvexHull(points)
#        vertices = points[hull.vertices]
#    else:

    # fall-back to Delaunay to support scipy 0.9 to 0.11
    # the documentation says:
    # "Computing convex hulls via the Delaunay triangulation is
    # inefficient and subject to increased numerical instability."
    hull = Delaunay(points).convex_hull
    vertices = points[np.unique(hull)]

    # sort vertices
    verticesCentered = vertices - vertices.mean(axis=0)
    angles = np.arctan2(verticesCentered[:,0], verticesCentered[:,1])
    vertices = vertices[np.argsort(angles)]
    
    return vertices

def findNearest(a, x):
    """
    Find item in sorted list a that is closest to x and return its index.
    
    If x has the same distance to both neighbors, the index of the left
    one is returned.
    """
    i = bisect.bisect_left(a, x)
    if i == len(a):
        return i-1
    else:
        if a[i] == x or i == 0:
            return i
        d0 = abs(x - a[i-1])
        d1 = abs(x - a[i])
        return i-1 if d0<=d1 else i

def extend(instance, new_class):
    """
    Apply inheritance after object creation.
    
    :see: https://stackoverflow.com/a/8545287
    """
    # note that new_class and instance.__class__ is switched here!
    instance.__class__ = type(
                '%s_extended_with_%s' % (instance.__class__.__name__, new_class.__name__), 
                (new_class, instance.__class__), 
                {}
            )



# Copyright European Space Agency, 2013

"""
Downloads a dataset from one of the following providers:

- ESA ISS Archive
- THEMIS Archive
"""

from __future__ import absolute_import, print_function
import argparse
from datetime import datetime
import os
import sys

__all__ = ['main']

def date(datestring):
    return datetime.strptime(datestring, '%Y-%m-%dT%H:%M:%S')

class Providers(object):
    ESA_ISS = 'esa-iss'
    THEMIS = 'themis'

description = '''
This tool downloads data of a given period from a provider and
stores it on disk without applying any postprocessing.
'''

epilog = '''
Examples:

auromat-download esa-iss --id 5
auromat-download esa-iss --id 5 --start 2000-01-01T12:00:00
auromat-download esa-iss --id 5 --id 6 --id 7
auromat-download esa-iss --id 5 --dir sequences/iss

auromat-download themis --start 2000-01-01T12:00:00 --end 2000-01-01T12:10:00
'''

def getParser():
    parser = argparse.ArgumentParser(prog='auromat-download', 
                                     epilog=epilog, description=description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('provider', help='The provider to get data from', 
                        choices=[Providers.ESA_ISS, Providers.THEMIS])
    period = parser.add_argument_group('period', 
                                       'These arguments specify which data to download.\n'
                                       'For the THEMIS provider, --start and --end are required.\n'
                                       'For the ESA ISS provider, --id is required and \n'
                                       '--start --end can be used to further constrain the period.')
    period.add_argument('--start', help='UTC start date, format 2000-01-01T12:00:00', type=date)
    period.add_argument('--end', help='UTC end date (inclusive)', type=date)
    period.add_argument('--id', help='Sequence ID(s) in ESA ISS archive', type=int, action='append')
    
    parser.add_argument('--dir', help='Directory to store data in, by default the current one', default=os.getcwd())
    parser.add_argument('--version', action='version', version='auromat TODO')
    return parser

def parseargs():
    parser = getParser()
    
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()
    
    if args.provider == Providers.THEMIS:
        if not (args.start and args.end):
            parser.error('--start and --end are required for the THEMIS provider')
    
    if args.provider == Providers.ESA_ISS:
        if not args.id:
            parser.error('--id is required for the ESA ISS provider')
    return args

def main():
    args = parseargs()
    
    if args.provider == Providers.ESA_ISS:
        from auromat.mapping.iss import ISSMappingProvider
        providers = [ISSMappingProvider(cacheFolder=os.path.join(args.dir, 'seq-{}'.format(id_)),
                                        id_=id_)
                     for id_ in args.id]
    
    elif args.provider == Providers.THEMIS:
        from auromat.mapping.themis import ThemisMappingProvider
        providers = [ThemisMappingProvider(cdfL1CacheFolder=args.dir, 
                                           cdfL2CacheFolder=args.dir)]
    
    for provider in providers:
        provider.download(args.start, args.end)
    
    print('Done.')
    
main.__doc__ = """
::
  
  {}
  

""".format(getParser().format_help().replace('\n', '\n  '))
    
if __name__ == '__main__':
    main()
    


# Copyright European Space Agency, 2013

"""
This module contains functions for calculating and working with
geodetic coordinates.
"""

from __future__ import division
from six.moves import range

import logging
from collections import namedtuple

import numpy as np
from geographiclib.geodesic import Geodesic
from geographiclib.constants import Constants


wgs84A = Constants.WGS84_a/1000
wgs84B = wgs84A * (1 - Constants.WGS84_f)

Location = namedtuple('Location', ['lat','lon']) # in degrees

def distance(location1, location2):
    """
    Return the shortest distance in meters between two locations.
    """
    data = Geodesic.WGS84.Inverse(location1.lat, location1.lon, 
                                  location2.lat, location2.lon, 
                                  Geodesic.DISTANCE)
    d = data['s12']
    return d

def angularDistance(location1, location2):
    """
    Return the shortest angular distance in degrees
    on an auxiliary sphere between two locations.
    """
    data = Geodesic.WGS84.Inverse(location1.lat, location1.lon, 
                                  location2.lat, location2.lon, 
                                  Geodesic.EMPTY)
    a = data['a12']
    return a

def line(location1, location2, resolution=1000):
    """
    Return points on the given geodesic in the given resolution.
    If the distance between the two locations is smaller than
    the requested resolution, then the line points consist only
    of the unchanged start and end locations.
    
    :type location1: Location
    :type location2: Location
    :param resolution: in meters
    :rtype: ndarray of shape (n, 2) with [lat,lon] order
    :return: line points in degrees
    """
    data = Geodesic.WGS84.Inverse(location1.lat, location1.lon, 
                                  location2.lat, location2.lon, 
                                  Geodesic.AZIMUTH | Geodesic.DISTANCE)
    az = data['azi1']
    d = data['s12']
    line = Geodesic.WGS84.Line(location1.lat, location1.lon, az, 
                               Geodesic.LATITUDE | Geodesic.LONGITUDE)
    
    num = d//resolution
    if num < 2:
        logging.warn('Geodesic line has less than two points at ' + str(resolution) + 'm resolution' +
                     'for a line length of ' + str(d) + 'm. The input points are returned as-is.')
        return np.array([[location1.lat, location1.lon], [location2.lat, location2.lon]])
    
    latLons = []
    ds = np.linspace(0, d, num)
    for dStartToPoint in ds:
        data = line.Position(dStartToPoint, Geodesic.LATITUDE | Geodesic.LONGITUDE)
        latLons.append([data['lat2'], data['lon2']])
    return np.array(latLons)

def destination(location, azimuth, distance):
    """
    Return the location when starting at `location` and
    travelling in direction `azimuth` for `distance` meters.
    
    :param lon, lat, azimuth: in degrees
    :param distance: in meters
    :rtype: :class:`Location`
    """
    data = Geodesic.WGS84.Direct(location.lat, location.lon, azimuth, distance)
    lon, lat = data['lon2'], data['lat2']
    return Location(lat, lon)

def intermediate(location1, location2, f=0.5):
    """
    Return the location when travelling `f * distance(location1,location2)` meters
    starting at `location1` in direction of `location2`.
    
    :type location1: Location
    :type location2: Location
    :param f: fraction to travel from `location1` to `location2`
    :rtype: :class:`Location`
    """
    data = Geodesic.WGS84.Inverse(location1.lat, location1.lon, 
                                  location2.lat, location2.lon, 
                                  Geodesic.AZIMUTH | Geodesic.DISTANCE)
    d = data['s12']
    az = data['azi1']
    data2 = Geodesic.WGS84.Direct(location1.lat, location1.lon, az, d*f, Geodesic.LATITUDE | Geodesic.LONGITUDE)
    lon, lat = data2['lon2'], data2['lat2']
    return Location(lat, lon)
    
def course(location1, location2):
    """
    Return the course/azimuth in degrees when travelling from `location1` to `location2`.
    """
    data = Geodesic.WGS84.Inverse(location1.lat, location1.lon, 
                                  location2.lat, location2.lon, 
                                  Geodesic.AZIMUTH)
    az = data['azi1']
    return az

def _courseDelta(a1, a2):
    """
    angles in degrees
    see http://www.element84.com/follow-up-to-determining-if-a-spherical-polygon-contains-a-pole.html
    """
    if a2 < a1:
        a2 += 360

    left_turn_amount = a2 - a1

    if left_turn_amount == 180:
        return 0
    elif left_turn_amount > 180:
        return left_turn_amount - 360
    else:
        return left_turn_amount
    
def _courseDeltaSum(points):
    """
    returns delta sum in integer degrees (either -360, -180, 0, 180, or 360)
    
    see:
    http://www.element84.com/determining-if-a-spherical-polygon-contains-a-pole.html
    http://www.element84.com/follow-up-to-determining-if-a-spherical-polygon-contains-a-pole.html
    
    Note: points must be in degrees as lat, lon pairs
    Note: points must be ordered such that they form a non-intersecting polygon
    Note: the first point must not be repeated at the end
    """
    # calculate initial and ending courses for each arc
    points = np.asarray(points)
    assert points.ndim == 2 and points.shape[1] == 2
    
    arcs = len(points)-1
    courses = np.empty(arcs*2)
    
    for i in range(arcs):
        lon1, lat1 = points[i,1], points[i,0]
        lon2, lat2 = points[i+1,1], points[i+1,0]
        courses[2*i] = course(Location(lat1,lon1), Location(lat2,lon2))
        courses[2*i+1] = course(Location(lat2,lon2), Location(lat1,lon1)) + 180

    # calculate course deltas
    deltas = np.empty(arcs*2)
    deltas[0] = _courseDelta(courses[arcs*2-1], courses[0])
    for i in range(1, arcs*2):
        deltas[i] = _courseDelta(courses[i-1], courses[i])
    
    deltaSum = np.around(np.sum(deltas), decimals=1)
    assert deltaSum in [-360,-180,0,180,360]
    return deltaSum

def containsOrCrossesPole(points):
    """
    Return whether the given polygon contains or crosses one of the poles.
    
    :param points: ordered points forming a non-intersecting unclosed polygon
    :type points: ndarray of shape (n,2) with lat,lon coordinates in degrees
    :rtype: bool 
    """
    deltaSum = _courseDeltaSum(points)

    containsPole = False
    if abs(deltaSum) == 360:
        logging.debug("pole: no")
    elif abs(deltaSum) == 180:
        containsPole = True
        logging.debug("pole: CROSSED")
    elif deltaSum == 0:
        containsPole = True
        logging.debug("pole: CONTAINED")
    return containsPole



# Copyright European Space Agency, 2013

from __future__ import division
import numpy as np
from auromat.utils import vectorLengths

try:
    from numexpr import evaluate as ne
except ImportError:
    ne = None

def sphereLineIntersection(sphereRadius, lineOrigin, lineDirection, directed=True):
    """
    Return the sphere-line intersection points.
            
    :param sphereRadius: radius of sphere with origin [0,0,0]
    :param lineOrigin: point, e.g. [1,2,3]
    :param lineDirection: unit vector or array of unit vectors
    :param bool directed: 
        True, if the line should be directed. In that case, the first intersection
        along the line is returned.
        If False, then the line is infinite in both ends and the intersection which
        is closest to the line origin is returned.
    :rtype: vector or array of vectors
    """
    directionPosDot = np.dot(lineDirection,lineOrigin)
    rootTerm = np.square(directionPosDot)
    rootTerm -= np.dot(lineOrigin,lineOrigin)
    rootTerm += np.square(sphereRadius)
    with np.errstate(invalid='ignore'): # ignore warnings for negative numbers (= no intersection)
        root = np.sqrt(rootTerm)
    directionPosDotNeg = -directionPosDot
    
    if directed:
        isInside = vectorLengths([lineOrigin])[0] < sphereRadius
        if isInside:
            d2 = directionPosDotNeg + root
            dMin = d2
        else:
            d1 = directionPosDotNeg - root
            dMin = d1
        dMin = _filterPointsOutsideDirectedLine(dMin)
    else:
        d1 = directionPosDotNeg - root
        d2 = directionPosDotNeg + root
        dMin = _closestDistance(d1, d2)

    return lineOrigin + dMin[...,None]*lineDirection

def _filterPointsOutsideDirectedLine(d):
    if d.ndim == 0:
        d = np.array(np.nan) if d < 0 else d
    else:
        with np.errstate(invalid='ignore'):
            d[d<0] = np.nan
    return d

def _ellipsoidLineIntersection_np(a, b, lineOrigin, lineDirection, directed=True):
    lineOrigin = np.require(lineOrigin, dtype=np.float64)
    lineDirection = np.require(lineDirection, dtype=np.float64)
    
    # turn into column vectors
    direction = lineDirection.T
    origin = -lineOrigin[:,None]
    
    radius = np.array([[1/a], [1/a], [1/b]])
    directionTimesRadius = direction * radius
    originTimesRadius = origin * radius
    
    # einsum is a bit faster for calculating the element-wise dot product
    # equivalent to np.sum(x*y, axis=0) or inner1d(x.T, y.T)
    directionDotOrigin = np.einsum("ij,ij->j", directionTimesRadius, originTimesRadius)
    directionDotDirection = np.einsum("ij,ij->j", directionTimesRadius, directionTimesRadius)
    originDotOrigin = np.einsum("ij,ij->j", originTimesRadius, originTimesRadius)

    rootTerm = np.square(directionDotOrigin)
    rootTerm -= originDotOrigin*directionDotDirection
    rootTerm += directionDotDirection
    with np.errstate(invalid='ignore'): # ignore warnings for negative numbers (= no intersection)
        np.sqrt(rootTerm, rootTerm)
    root = rootTerm

    if directed:
        if _isInsideEllipsoid(lineOrigin, a, b):
            d2 = directionDotOrigin
            d2 += root
            dMin = d2
        else:
            d1 = directionDotOrigin
            d1 -= root
            dMin = d1
        dMin = _filterPointsOutsideDirectedLine(dMin)
    else:
        d1 = directionDotOrigin - root
        d2 = directionDotOrigin
        d2 += root
        dMin = _closestDistance(d1, d2)
    
    dMin /= directionDotDirection
    
    res = directionTimesRadius
    np.multiply(direction, dMin, res)
    res -= origin
    return res.T

def _ellipsoidLineIntersection_ne(a, b, lineOrigin, lineDirection, directed=True):
    lineOrigin = np.require(lineOrigin, dtype=np.float64)
    lineDirection = np.require(lineDirection, dtype=np.float64)
    
    # turn into column vectors
    direction = lineDirection.T
    origin = -lineOrigin[:,None]
    
    radius = np.array([[1/a], [1/a], [1/b]])
    directionTimesRadius = ne('direction * radius')
    originTimesRadius = ne('origin * radius')
    
    directionDotOrigin = np.einsum("ij,ij->j", directionTimesRadius, originTimesRadius)
    directionDotDirection = np.einsum("ij,ij->j", directionTimesRadius, directionTimesRadius)
    originDotOrigin = np.einsum("ij,ij->j", originTimesRadius, originTimesRadius)
        
    root = ne('sqrt(directionDotOrigin**2 - originDotOrigin*directionDotDirection + directionDotDirection)')
        
    if directed:
        if _isInsideEllipsoid(lineOrigin, a, b):
            d2 = directionDotOrigin
            ne('directionDotOrigin + root', out=d2)
            dMin = d2
        else:
            d1 = directionDotOrigin
            ne('directionDotOrigin - root', out=d1)
            dMin = d1
        dMin = _filterPointsOutsideDirectedLine(dMin)
    else:
        d1 = ne('directionDotOrigin - root')
        d2 = directionDotOrigin
        ne('directionDotOrigin + root', out=d2)
        dMin = _closestDistance(d1, d2)
    
    res = directionTimesRadius
    ne('direction * (dMin / directionDotDirection) - origin', out=res)
    return res.T

def ellipsoidLineIntersection(a, b, lineOrigin, lineDirection, directed=True):
    """
    Return the ellipsoid-line intersection points.

    :note: The ellipsoid is assumed to be at (0,0,0).
    :param a: equatorial axis of the ellipsoid of revolution
    :param b: polar axis of the ellipsoid of revolution
    :param lineOrigin: x,y,z vector
    :param lineDirection: x,y,z array of vectors; not required to be unit vectors
    :param bool directed: 
        True, if the line should be directed. In that case, the first intersection
        along the line is returned.
        If False, then the line is infinite in both ends and the intersection which
        is closest to the line origin is returned.
    :rtype: vector or array of vectors
    """
    if ne:
        return _ellipsoidLineIntersection_ne(a, b, lineOrigin, lineDirection, directed)
    else:
        return _ellipsoidLineIntersection_np(a, b, lineOrigin, lineDirection, directed)

def _ellipsoidLineIntersects_np(a, b, lineOrigin, lineDirection, directed=True):
    lineOrigin = np.require(lineOrigin, dtype=np.float64)
    lineDirection = np.require(lineDirection, dtype=np.float64)
    
    # turn into column vectors
    direction = lineDirection.T
    origin = -lineOrigin[:,None]
    
    radius = np.array([[1/a], [1/a], [1/b]])
        
    directionTimesRadius = direction * radius
    originTimesRadius = origin * radius
    directionDotOrigin = np.einsum("ij,ij->j", directionTimesRadius, originTimesRadius)
    directionDotDirection = np.einsum("ij,ij->j", directionTimesRadius, directionTimesRadius)
    originDotOrigin = np.einsum("ij,ij->j", originTimesRadius, originTimesRadius)

    rootTerm = np.square(directionDotOrigin)
    rootTerm -= originDotOrigin*directionDotDirection
    rootTerm += directionDotDirection
    
    with np.errstate(invalid='ignore'): # ignore warnings for negative numbers (= no intersection)
        if directed:
            np.sqrt(rootTerm, rootTerm)
            root = rootTerm
            if _isInsideEllipsoid(lineOrigin, a, b):
                d2 = directionDotOrigin
                d2 += root
                dMin = d2
            else:
                d1 = directionDotOrigin
                d1 -= root
                dMin = d1
            intersects = dMin >= 0
        else:
            intersects = rootTerm >= 0
    
    return intersects

def _ellipsoidLineIntersects_ne(a, b, lineOrigin, lineDirection, directed=True):
    lineOrigin = np.require(lineOrigin, dtype=np.float64)
    lineDirection = np.require(lineDirection, dtype=np.float64)
    
    # turn into column vectors
    direction = lineDirection.T
    origin = -lineOrigin[:,None]
    
    radius = np.array([[1/a], [1/a], [1/b]])
    directionTimesRadius = ne('direction * radius')
    originTimesRadius = ne('origin * radius')
    
    dirDotOri = np.einsum("ij,ij->j", directionTimesRadius, originTimesRadius)
    dirDotDir = np.einsum("ij,ij->j", directionTimesRadius, directionTimesRadius)
    oriDotOri = np.einsum("ij,ij->j", originTimesRadius, originTimesRadius)

    if directed:
        if _isInsideEllipsoid(lineOrigin, a, b):
            intersects = ne('dirDotOri + sqrt(dirDotOri**2 - oriDotOri*dirDotDir + dirDotDir) >= 0')
        else:
            intersects = ne('dirDotOri - sqrt(dirDotOri**2 - oriDotOri*dirDotDir + dirDotDir) >= 0')
    else:
        intersects = ne('dirDotOri**2 - oriDotOri*dirDotDir + dirDotDir >= 0')
        
    return intersects

def ellipsoidLineIntersects(a, b, lineOrigin, lineDirection, directed=True):
    """
    As :func:`ellipsoidLineIntersection` but returns an array of booleans instead
    of the intersection points.
    """
    if ne:
        return _ellipsoidLineIntersects_ne(a, b, lineOrigin, lineDirection, directed)
    else:
        return _ellipsoidLineIntersects_np(a, b, lineOrigin, lineDirection, directed)

def _isInsideEllipsoid(point, a, b):
    x,y,z = point
    return (x/a)**2 + (y/a)**2 + (z/b)**2 < 1

def _closestDistance(d1, d2):
    """
    Assuming a single line origin, this returns the distances (either d1 or d2)
    whose absolute values are smallest.
    """
    with np.errstate(invalid='ignore'):
        dMin = np.where(np.abs(d1) < np.abs(d2), d1, d2)
    return dMin



# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

import warnings
import calendar
import datetime

import numpy as np

from astropy import units
from astropy import coordinates as coord

import ephem

class EphemerisCalculator:
    """
    Calculates J2000 positions using a set of TLEs for a *single* satellite.
    TLEs are read once into a string to serve as a cache.
    
    Use the :mod:`~auromat.coordinates.spacetrack` module for downloading
    TLE data.
    
    Note that this module assumes continuous TLE data as created by
    the :mod:`~auromat.coordinates.spacetrack` module. 
    
    Example::
    
      from datetime import datetime
      from auromat.coordinates.spacetrack import Spacetrack
      from auromat.coordinates.ephem import EphemerisCalculator
      path = 'iss_tles.txt'
      st = Spacetrack('user', 'pass')
      st.updateTLEs(25544, path)
      iss = EphemerisCalculator(path)
      j2000_xyz = iss(datetime(2012,1,1,12,30,0))
      print(j2000_xyz)
      
    """
    def __init__(self, tleFilePath):
        """
        :param string tleFilePath: path to file containing TLE sets
        """
        with open(tleFilePath, 'r') as t:
            self.tles = t.readlines()
                
        assert len(self.tles) > 0
        assert len(self.tles) % 2 == 0
         
    def __call__(self, date):
        """
        Return the J2000 position of the space object for the given date.
        
        :param datetime.datetime date:
        :return: x,y,z position as array
        :rtype: ndarray of shape (3,)
        """
        return self.getPosition(date)
    
    @property
    def noradId(self):
        """
        The NORAD ID as read from the first TLE.
        
        :rtype: str
        """
        return self.tles[0][2:7]

    @property
    def firstEpoch(self):
        """
        The earliest epoch in the TLE file.
        
        :rtype: datetime.datetime
        """
        tle = ephem.readtle('foo', self.tles[0], self.tles[1])
        date = self._toDateTime(tle._epoch)
        return date
    
    @property
    def lastEpoch(self):
        """
        The latest epoch in the TLE file.
        
        :rtype: datetime.datetime
        """
        tle = ephem.readtle('foo', self.tles[-2], self.tles[-1])
        date = self._toDateTime(tle._epoch)
        return date
    
    def contains(self, date):
        """
        Return whether firstEpoch <= date <= lastEpoch holds. 
        
        :type date: datetime.datetime
        """
        return self.firstEpoch <= date <= self.lastEpoch
    
    @staticmethod
    def _toDateTime(ephemDate):
        """
        Converts PyEphem dates to datetime objects.
        """
        timestamp = calendar.timegm(ephemDate.tuple())
        return datetime.datetime.utcfromtimestamp(timestamp)
    
    def getTLE(self, date):
        date = ephem.Date(date)
        
        tleCount = len(self.tles) // 2
        lo = 0
        hi = tleCount
        while lo < hi: # equals bisect.bisect_left
            mid = (lo+hi)//2
            tle = ephem.readtle('foo', self.tles[mid*2], self.tles[mid*2 + 1])
            if tle._epoch < date:
                lo = mid+1
            else:
                hi = mid

        tleIdx = lo-1 # equals find_lt (rightmost TLE with epoch less than given date)

        return self.tles[tleIdx*2], self.tles[tleIdx*2 + 1]
       
    def getPosition(self, date):
        """
        Return the J2000 position of the space object for the given date.
        
        Note that you can also use the call operator::
        
          iss = EphemerisCalculator(path)
          iss(date)
        
        :type date: datetime.datetime
        :rtype: ndarray of xyz J2000 coordinates in km
        """
        tlelines = self.getTLE(date)
        tle = ephem.readtle('foo', tlelines[0], tlelines[1])
        
        date = ephem.Date(date)
        
        if tle._epoch > date:
            raise Exception("The epoch of the earliest available TLE is AFTER the requested date. " +
                            "Are you missing historic TLE data?")
        
        if (date - tle._epoch > 24*ephem.hour):
            warnings.warn('closest TLE epoch is ' + str((date - tle._epoch)*24) + 'h away from photo time')                          
                
        tle.compute(date)
        
        # see http://stackoverflow.com/q/19426505/60982
        earthRadiusPyEphem = 6378.16 * units.km
        distanceEarthCenter = tle.elevation*units.m + earthRadiusPyEphem
        
        xyz = coord.distances.spherical_to_cartesian(distanceEarthCenter.to(units.km).value, tle.a_dec, tle.a_ra)
        return np.array(xyz)
        
    def _getPositionX(self, date):
        """
        experimental: uses sgp4 library
        
        NOTE: returns TEME coordinates (true equator mean equinox)

        :param datetime date:
        :rtype: GCRSCoordinates
        """
        from sgp4.earth_gravity import wgs72
        from sgp4.io import twoline2rv
        from sgp4.ext import jday

        jdate = jday(date.year, date.month, date.day, date.hour, date.minute, date.second)

        tleCount = len(self.tles) // 2
        lo = 0
        hi = tleCount
        while lo < hi: # equals bisect.bisect_left
            mid = (lo+hi)//2
            tle = twoline2rv(self.tles[mid*2], self.tles[mid*2 + 1], wgs72)
            if tle.jdsatepoch < jdate:
                lo = mid+1
            else:
                hi = mid
        
        tleIdx = lo-1 # equals find_lt (rightmost TLE with epoch less than given date)
        tle = twoline2rv(self.tles[tleIdx*2], self.tles[tleIdx*2 + 1], wgs72)

        if tle.jdsatepoch > jdate:
            raise Exception("The epoch of the earliest available TLE is AFTER the requested date. " +
                            "Are you missing historic TLE data?")
        
        if (jdate - tle.jdsatepoch > 10*ephem.hour):
            warnings.warn('closest TLE epoch is ' + str((date - tle.jdsatepoch)*24) + 'h away from photo time')
        
        position, _ = tle.propagate(date.year, date.month, date.day, date.hour, date.minute, date.second)

        x,y,z = units.km.to(units.pc, position)
        
        # FIXME this is wrong, conversion TEME-J2000 is missing
        # -> need skyfield library for that, but too heavy currently
        #   see https://github.com/brandon-rhodes/python-skyfield/issues/31
        return np.array([x,y,z]) 


# Copyright (c) 2005-2011, NumPy Developers.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
#     * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
# 
#     * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
# 
#     * Neither the name of the NumPy Developers nor the names of any
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module contains a fast replacement for numpy's histogramdd and histogram2d.
Two changes were made. The first was replacing

np.digitize(a, b)

with 

np.searchsorted(b, a, "right")

This performance bug is explained on https://github.com/numpy/numpy/issues/2656
The speedup is around 2x for big number of bins (roughly >100).
It assumes that the bins are monotonic.

The other change is to allow lists of weight arrays. This is advantageous for
resampling as there is just one set of coordinates but several data arrays (=weights).
Therefore repeated computations are prevented.
"""


import numpy as np
from numpy import atleast_2d, asarray, zeros, ones, array, atleast_1d, arange,\
    isscalar, linspace, diff, empty, around, where, bincount, sort, log10,\
    searchsorted

def histogramdd(sample, bins=10, range=None, normed=False, weights=None):
    """
    Compute the multidimensional histogram of some data.

    Parameters
    ----------
    sample : array_like
        The data to be histogrammed. It must be an (N,D) array or data
        that can be converted to such. The rows of the resulting array
        are the coordinates of points in a D dimensional polytope.
    bins : sequence or int, optional
        The bin specification:

        * A sequence of arrays describing the bin edges along each dimension.
        * The number of bins for each dimension (nx, ny, ... =bins)
        * The number of bins for all dimensions (nx=ny=...=bins).

    range : sequence, optional
        A sequence of lower and upper bin edges to be used if the edges are
        not given explicitly in `bins`. Defaults to the minimum and maximum
        values along each dimension.
    normed : bool, optional
        If False, returns the number of samples in each bin. If True,
        returns the bin density ``bin_count / sample_count / bin_volume``.
    weights : array_like (N,), optional
        An array of values `w_i` weighing each sample `(x_i, y_i, z_i, ...)`.
        Weights are normalized to 1 if normed is True. If normed is False,
        the values of the returned histogram are equal to the sum of the
        weights belonging to the samples falling into each bin.
        Weights can also be a list of (weight arrays or None), in which case
        a list of histograms is returned as H.

    Returns
    -------
    H : ndarray
        The multidimensional histogram of sample x. See normed and weights
        for the different possible semantics.
    edges : list
        A list of D arrays describing the bin edges for each dimension.

    See Also
    --------
    histogram: 1-D histogram
    histogram2d: 2-D histogram

    Examples
    --------
    >>> r = np.random.randn(100,3)
    >>> H, edges = np.histogramdd(r, bins = (5, 8, 4))
    >>> H.shape, edges[0].size, edges[1].size, edges[2].size
    ((5, 8, 4), 6, 9, 5)

    """

    try:
        # Sample is an ND-array.
        N, D = sample.shape
    except (AttributeError, ValueError):
        # Sample is a sequence of 1D arrays.
        sample = atleast_2d(sample).T
        N, D = sample.shape
    
    if weights is None:
        W = None
    else:    
        try:
            # Weights is a 1D-array
            weights.shape
            W = -1
        except (AttributeError, ValueError):
            # Weights is a list of 1D-arrays or None's
            W = len(weights)

    if W == -1 and weights.ndim != 1:
        raise AttributeError('Weights must be a 1D-array, None, or a list of both')

    nbin = empty(D, int)
    edges = D*[None]
    dedges = D*[None]
    if weights is not None:
        if W == -1:
            weights = asarray(weights)
            assert weights.shape == (N,)
        else:
            for i in arange(W):
                if weights[i] is not None:
                    weights[i] = asarray(weights[i])
                    assert weights[i].shape == (N,)

    try:
        M = len(bins)
        if M != D:
            raise AttributeError(
                'The dimension of bins must be equal to the dimension of the '
                ' sample x.')
    except TypeError:
        # bins is an integer
        bins = D*[bins]

    # Select range for each dimension
    # Used only if number of bins is given.
    if range is None:
        # Handle empty input. Range can't be determined in that case, use 0-1.
        if N == 0:
            smin = zeros(D)
            smax = ones(D)
        else:
            smin = atleast_1d(array(sample.min(0), float))
            smax = atleast_1d(array(sample.max(0), float))
    else:
        smin = zeros(D)
        smax = zeros(D)
        for i in arange(D):
            smin[i], smax[i] = range[i]

    # Make sure the bins have a finite width.
    for i in arange(len(smin)):
        if smin[i] == smax[i]:
            smin[i] = smin[i] - .5
            smax[i] = smax[i] + .5

    # Create edge arrays
    for i in arange(D):
        if isscalar(bins[i]):
            if bins[i] < 1:
                raise ValueError(
                    "Element at index %s in `bins` should be a positive "
                    "integer." % i)
            nbin[i] = bins[i] + 2  # +2 for outlier bins
            edges[i] = linspace(smin[i], smax[i], nbin[i]-1)
        else:
            edges[i] = asarray(bins[i], float)
            nbin[i] = len(edges[i]) + 1  # +1 for outlier bins
        dedges[i] = diff(edges[i])
        if np.any(np.asarray(dedges[i]) <= 0):
            raise ValueError(
                "Found bin edge of size <= 0. Did you specify `bins` with"
                "non-monotonic sequence?")

    nbin = asarray(nbin)

    # Handle empty input.
    if N == 0:
        if W > 0:
            return [np.zeros(nbin-2) for _ in arange(W)], edges
        else:
            return np.zeros(nbin-2), edges

    # Compute the bin number each sample falls into.
    Ncount = {}
    for i in arange(D):
        # searchsorted is faster for many bins
        Ncount[i] = searchsorted(edges[i], sample[:, i], "right")
        #Ncount[i] = digitize(sample[:, i], edges[i])

    # Using digitize, values that fall on an edge are put in the right bin.
    # For the rightmost bin, we want values equal to the right
    # edge to be counted in the last bin, and not as an outlier.
    for i in arange(D):
        # Rounding precision
        mindiff = dedges[i].min()
        if not np.isinf(mindiff):
            decimal = int(-log10(mindiff)) + 6
            # Find which points are on the rightmost edge.
            not_smaller_than_edge = (sample[:, i] >= edges[i][-1])
            on_edge = (around(sample[:, i], decimal) == around(edges[i][-1], decimal))
            # Shift these points one bin to the left.
            Ncount[i][where(on_edge & not_smaller_than_edge)[0]] -= 1

    # Compute the sample indices in the flattened histogram matrix.
    ni = nbin.argsort()
    xy = zeros(N, int)
    for i in arange(0, D-1):
        xy += Ncount[ni[i]] * nbin[ni[i+1:]].prod()
    xy += Ncount[ni[-1]]

    # Compute the number of repetitions in xy and assign it to the
    # flattened histmat.
    if len(xy) == 0:
        if W > 0:
            return [np.zeros(nbin-2) for _ in arange(W)], edges
        else:
            return zeros(nbin-2, int), edges

    # Flattened histogram matrix (1D)
    # Reshape is used so that overlarge arrays
    # will raise an error.
    Wd = W if W > 0 else 1
    hists = [zeros(nbin, float).reshape(-1) for _ in arange(Wd)]
    for histidx, hist in enumerate(hists):
        weights_ = weights[histidx] if W > 0 else weights
        flatcount = bincount(xy, weights_)
        a = arange(len(flatcount))
        hist[a] = flatcount
    
        # Shape into a proper matrix
        hist = hist.reshape(sort(nbin))
        ni = nbin.argsort()
        for i in arange(nbin.size):
            j = ni.argsort()[i]
            hist = hist.swapaxes(i, j)
            ni[i], ni[j] = ni[j], ni[i]
    
        # Remove outliers (indices 0 and -1 for each dimension).
        core = D*[slice(1, -1)]
        hist = hist[core]
    
        # Normalize if normed is True
        if normed:
            s = hist.sum()
            for i in arange(D):
                shape = ones(D, int)
                shape[i] = nbin[i] - 2
                hist = hist / dedges[i].reshape(shape)
            hist /= s
    
        if (hist.shape != nbin - 2).any():
            raise RuntimeError(
                "Internal Shape Error: hist.shape != nbin-2 -> " + str(hist.shape) + " != " + str(nbin-2))
        
        hists[histidx] = hist
    
    if W in [None, -1]:
        return hists[0], edges
    else:
        return hists, edges

def histogram2d(x, y, bins=10, range=None, normed=False, weights=None):
    """
    Compute the bi-dimensional histogram of two data samples.

    Parameters
    ----------
    x : array_like, shape (N,)
        An array containing the x coordinates of the points to be
        histogrammed.
    y : array_like, shape (N,)
        An array containing the y coordinates of the points to be
        histogrammed.
    bins : int or [int, int] or array_like or [array, array], optional
        The bin specification:

          * If int, the number of bins for the two dimensions (nx=ny=bins).
          * If [int, int], the number of bins in each dimension
            (nx, ny = bins).
          * If array_like, the bin edges for the two dimensions
            (x_edges=y_edges=bins).
          * If [array, array], the bin edges in each dimension
            (x_edges, y_edges = bins).

    range : array_like, shape(2,2), optional
        The leftmost and rightmost edges of the bins along each dimension
        (if not specified explicitly in the `bins` parameters):
        ``[[xmin, xmax], [ymin, ymax]]``. All values outside of this range
        will be considered outliers and not tallied in the histogram.
    normed : bool, optional
        If False, returns the number of samples in each bin. If True,
        returns the bin density ``bin_count / sample_count / bin_area``.
    weights : array_like, shape(N,), optional
        An array of values ``w_i`` weighing each sample ``(x_i, y_i)``.
        Weights are normalized to 1 if `normed` is True. If `normed` is
        False, the values of the returned histogram are equal to the sum of
        the weights belonging to the samples falling into each bin.
        Weights can also be a list of (weight arrays or None), in which case
        a list of histograms is returned as H.

    Returns
    -------
    H : ndarray, shape(nx, ny)
        The bi-dimensional histogram of samples `x` and `y`. Values in `x`
        are histogrammed along the first dimension and values in `y` are
        histogrammed along the second dimension.
    xedges : ndarray, shape(nx,)
        The bin edges along the first dimension.
    yedges : ndarray, shape(ny,)
        The bin edges along the second dimension.

    See Also
    --------
    histogram : 1D histogram
    histogramdd : Multidimensional histogram

    Notes
    -----
    When `normed` is True, then the returned histogram is the sample
    density, defined such that the sum over bins of the product
    ``bin_value * bin_area`` is 1.

    Please note that the histogram does not follow the Cartesian convention
    where `x` values are on the abscissa and `y` values on the ordinate
    axis.  Rather, `x` is histogrammed along the first dimension of the
    array (vertical), and `y` along the second dimension of the array
    (horizontal).  This ensures compatibility with `histogramdd`.

    Examples
    --------
    >>> import matplotlib as mpl
    >>> import matplotlib.pyplot as plt

    Construct a 2D-histogram with variable bin width. First define the bin
    edges:

    >>> xedges = [0, 1, 1.5, 3, 5]
    >>> yedges = [0, 2, 3, 4, 6]

    Next we create a histogram H with random bin content:

    >>> x = np.random.normal(3, 1, 100)
    >>> y = np.random.normal(1, 1, 100)
    >>> H, xedges, yedges = np.histogram2d(y, x, bins=(xedges, yedges))

    Or we fill the histogram H with a determined bin content:

    >>> H = np.ones((4, 4)).cumsum().reshape(4, 4)
    >>> print H[::-1]  # This shows the bin content in the order as plotted
    [[ 13.  14.  15.  16.]
     [  9.  10.  11.  12.]
     [  5.   6.   7.   8.]
     [  1.   2.   3.   4.]]

    Imshow can only do an equidistant representation of bins:

    >>> fig = plt.figure(figsize=(7, 3))
    >>> ax = fig.add_subplot(131)
    >>> ax.set_title('imshow: equidistant')
    >>> im = plt.imshow(H, interpolation='nearest', origin='low',
                    extent=[xedges[0], xedges[-1], yedges[0], yedges[-1]])

    pcolormesh can display exact bin edges:

    >>> ax = fig.add_subplot(132)
    >>> ax.set_title('pcolormesh: exact bin edges')
    >>> X, Y = np.meshgrid(xedges, yedges)
    >>> ax.pcolormesh(X, Y, H)
    >>> ax.set_aspect('equal')

    NonUniformImage displays exact bin edges with interpolation:

    >>> ax = fig.add_subplot(133)
    >>> ax.set_title('NonUniformImage: interpolated')
    >>> im = mpl.image.NonUniformImage(ax, interpolation='bilinear')
    >>> xcenters = xedges[:-1] + 0.5 * (xedges[1:] - xedges[:-1])
    >>> ycenters = yedges[:-1] + 0.5 * (yedges[1:] - yedges[:-1])
    >>> im.set_data(xcenters, ycenters, H)
    >>> ax.images.append(im)
    >>> ax.set_xlim(xedges[0], xedges[-1])
    >>> ax.set_ylim(yedges[0], yedges[-1])
    >>> ax.set_aspect('equal')
    >>> plt.show()

    """
    try:
        N = len(bins)
    except TypeError:
        N = 1

    if N != 1 and N != 2:
        xedges = yedges = asarray(bins, float)
        bins = [xedges, yedges]
    hist, edges = histogramdd([x, y], bins, range, normed, weights)
    return hist, edges[0], edges[1]


# Copyright European Space Agency, 2013

"""
This module resamples mappings in a given resolution in the plate carree
projection, relative to either geodetic or MLat/MLT coordinates.
"""

from __future__ import division, print_function, absolute_import

from six.moves import map

import time

import numpy as np
import numpy.ma as ma

from distutils.version import LooseVersion
import astropy
import copy
from functools import partial
if LooseVersion(astropy.__version__) < '0.4':
    raise RuntimeError('astropy<0.4 is unsupported due to bugs in handling non-contiguous arrays')
from astropy.coordinates import Angle
import astropy.units as u

import scipy.interpolate

from auromat.utils import pointsInsidePolygon, extend
from auromat.mapping.mapping import BaseMapping, MappingCollection,\
    convertMappingToSM, convertSMMappingToGeo
from auromat.coordinates.transform import rotatePole      
from auromat.util.histogram import histogram2d
from auromat.coordinates import geodesic
from auromat.coordinates.geodesic import Location

def plateCarreeResolution(boundingBox, arcsecPerPx):
    """
    Approximates the latitude and longitude resolution of a plate carree
    projection from the given spherical resolution for the area given
    by the bounding box. The approximation is calculated for the bounding box center.
    
    :type boundingBox: auromat.mapping.mapping.BoundingBox
    :param arcsecPerPx: spherical resolution
    :rtype: tuple (latPxPerDeg, lonPxPerDeg)
    """
    degPerPx = (arcsecPerPx * u.arcsec).to(u.degree).value
    latPxPerDeg = 1/degPerPx
    
    latMiddle = (boundingBox.latNorth + boundingBox.latSouth)/2
    middleLeft = Location(latMiddle, boundingBox.lonWest)
    middleRight = Location(latMiddle, boundingBox.lonEast)
    lonMiddleDistance = geodesic.angularDistance(middleLeft, middleRight)
    px = lonMiddleDistance/degPerPx
    lonEast = boundingBox.lonEast
    if boundingBox.lonWest > lonEast:
        lons = lonEast+360 - boundingBox.lonWest
    else:
        lons = lonEast - boundingBox.lonWest
    lonMiddlePxPerDeg = px/lons
    
    return latPxPerDeg, lonMiddlePxPerDeg    

def resampleMLatMLT(mapping, **kw):
    """ Resamples a mapping such that MLat/MLT become regular grids.
    
    See :func:`resample` for parameters.    
    """
    sm = convertMappingToSM(mapping)
    smResampled = resample(sm, **kw)
    geo = convertSMMappingToGeo(smResampled)
    return geo

def resample(mappingOrCollection, pxPerDeg=25, arcsecPerPx=None, containsPole=None, method='mean'):
    """
    Returns a new mapping (or collection) where the colors and elevation
    are resampled into a regular latitude/longitude grid (plate carree projection)
    with y=latitude and x=longitude.
    
    If 'mean' binning is used as resampling method then take into account that
    this will lead to holes for low elevation angles if a high resampling resolution
    is used. This is because binning does not interpolate when there are zero data
    points in a given bin. Mask the mapping by elevation (e.g. 10deg) to get rid
    of the areas with holes.
    
    :param mappingOrCollection:
    :param None|number|tuple pxPerDeg: tuple (latPxPerDeg, lonPxPerDeg) or a number if both are the same
    :param None|number arcsecPerPx: spherical resolution, used to approximate pxPerDeg,
                                    has precedence over pxPerDeg
    :param None|bool containsPole: specify True|False to skip pole checking algorithm
    :param method: binning: 'mean'; interpolation: 'nearest', 'linear', 'cubic';
                   Note that linear and cubic take considerably longer and use much more memory while
                   they don't bring any benefit over 'nearest' if the goal is downsampling.
    :rtype: a subclass of BaseMapping or MappingCollection
    """
    def doResample(mapping, pxPerDeg, arcsecPerPx, containsPole):
        # trigger calculation of properties so that they are not included in the timing measurements
        mapping.lats
        mapping.latsCenter
        mapping.elevation
        mapping.img
        t0 = time.time()
        
        if containsPole is None:
            containsPole = mapping.containsPole
            
        if arcsecPerPx:
            pxPerDeg = plateCarreeResolution(mapping.boundingBox, arcsecPerPx)
        else:
            try:
                _, _ = pxPerDeg
            except TypeError:
                assert pxPerDeg is not None
                pxPerDeg = (pxPerDeg, pxPerDeg)
        print('pxPerDeg: ' + str(pxPerDeg))
        
        imgIsInt = mapping.img.dtype in [np.uint8, np.uint16, np.uint32, np.uint64, np.int8, np.int16, np.int32, np.int64]
        
        # merge elevation with rgb array and extract channels afterwards
        merged = np.dstack((mapping.img.astype(np.float64).filled(np.nan),
                            mapping.elevation.filled(np.nan)))
        lats, lons, latsCenter, lonsCenter, merged = \
            _resample(mapping.latsCenter.filled(np.nan), mapping.lonsCenter.filled(np.nan), mapping.altitude,
                      merged,
                      lambda: mapping.outline, mapping.boundingBox,
                      pxPerDeg, mapping.containsDiscontinuity, containsPole,
                      method=method)
                
        img, elevation = np.dsplit(merged, [-1])
        if imgIsInt:
            with np.errstate(invalid='ignore'):
                img = np.round(img)
        img = np.require(ma.masked_invalid(img, copy=False), mapping.img.dtype)
        if mapping.img.ndim == 2:
            img = img.reshape(img.shape[0], img.shape[1])            
        elevation = elevation.reshape(elevation.shape[0],elevation.shape[1])
        elevation = ma.masked_invalid(elevation, copy=False)
    
        resampledMapping = mapping.createResampled(lats, lons, latsCenter, lonsCenter, elevation, img)
        
        print('resampling:', time.time()-t0, 's')
        
        return resampledMapping
    
    if isinstance(mappingOrCollection, BaseMapping):
        resampled = doResample(mappingOrCollection, pxPerDeg, arcsecPerPx, containsPole)
        
    elif isinstance(mappingOrCollection, MappingCollection):
        mappings = []
        for mapping in mappingOrCollection.mappings:
            mappings.append(doResample(mapping, pxPerDeg, arcsecPerPx, containsPole))
        resampled = MappingCollection(mappings, mayOverlap=mappingOrCollection.mayOverlap)
    
    else:
        raise ValueError('First argument must be a mapping or a mapping collection, but is: {}'.
                         format(type(mappingOrCollection)))
    
    return resampled   

def _resample(latsCenter, lonsCenter, altitude, data, outlineLatLonFn, boundingBox, pxPerDeg, 
              containsDiscontinuity=False, containsPole=False, method='mean'):
    """
    
    Note: Each channel is resampled on its own.

    :param latsCenter: (h,w)
    :param lonsCenter: (h,w)
    :param data: float data for each pixel center, (h,w,n) with n>0, or (h,w)
    :param pxPerDeg: tuple (latPxPerDeg, lonPxPerDeg)
    :rtype: tuple (lat, lon, latCenter, lonCenter, data)
    """   
    latMin = boundingBox.latSouth
    latMax = boundingBox.latNorth
    lonMin = boundingBox.lonWest
    lonMax = boundingBox.lonEast
        
    if containsPole:
        print('contains pole')
        
        outlineLatLon = outlineLatLonFn()
        outlineLats = outlineLatLon[:,0]
        outlineLons = outlineLatLon[:,1]
        
        # rotation of latitude/poles needs to happen in cartesian space based on earth as sphere
        # only a very small error will be introduced here as the outline form is not a segment of a sphere but an ellipsoid
        # -> as the outline is very small in size, this won't be a problem
        angle = 90
        axis = [1,0,0]
        outlineLats, outlineLons = rotatePole(np.deg2rad(outlineLats), np.deg2rad(outlineLons), 
                                              altitude, angle=angle, axis=axis)
        outlineLats, outlineLons = np.rad2deg(outlineLats), np.rad2deg(outlineLons)
        
        outlineLatLon[:,0] = outlineLats
        outlineLatLon[:,1] = outlineLons
        
        latMin, latMax = np.min(outlineLats), np.max(outlineLats)
        lonMin, lonMax = np.min(outlineLons), np.max(outlineLons)
                
        latsCenter_, lonsCenter_ = rotatePole(np.deg2rad(np.ravel(latsCenter)), np.deg2rad(np.ravel(lonsCenter)), 
                                              altitude, angle=angle, axis=axis)
        latsCenter = np.rad2deg(latsCenter_.reshape(latsCenter.shape))
        lonsCenter = np.rad2deg(lonsCenter_.reshape(lonsCenter.shape))
        
    elif containsDiscontinuity:
        print('contains discontinuity')
        
        outlineLatLon = outlineLatLonFn()
        outlineLats = outlineLatLon[:,0]
        outlineLons = outlineLatLon[:,1]

        # rotate longitudes out of 180° discontinuity; poles stay where they are
        # this introduces no additional error (e.g. due to ellipsoidal form)
        angle = 180
        outlineLons = Angle((outlineLons + angle) * u.deg).wrap_at(angle * u.deg).degree
        outlineLatLon[:,1] = outlineLons
            
        lonMin, lonMax = np.min(outlineLons), np.max(outlineLons)
        
        lonsCenter = Angle((lonsCenter + angle) * u.deg).wrap_at(angle * u.deg).degree

    # create regular plate carree grid within bounding box where y=lat and x=lon
    # Note: For a given pxPerDeg, all resamplings are aligned to the same global grid.
    latPxPerDeg, lonPxPerDeg = pxPerDeg
    assert latPxPerDeg > 0 and lonPxPerDeg > 0
    nLat, nLon, latMinInGrid, latMaxInGrid, lonMinInGrid, lonMaxInGrid =\
        fixedGrid(pxPerDeg, latMin, latMax, lonMin, lonMax)
    assert nLat > 1, 'nlat={}, latMax={}, latMin={}, pxperdeg={}'.format(nLat, latMaxInGrid, latMinInGrid, pxPerDeg)
    assert nLon > 1, 'nlon={}, lonMax={}, lonMin={}, pxperdeg={}'.format(nLon, lonMaxInGrid, lonMinInGrid, pxPerDeg)
    # the center coordinates are the ones which lie on the grid, corners are calculated
    latSpaceCenter, latStep = np.linspace(latMaxInGrid, latMinInGrid, num=nLat, retstep=True)
    lonSpaceCenter, lonStep = np.linspace(lonMinInGrid, lonMaxInGrid, num=nLon, retstep=True)
    
    # skip first and last coordinate center coordinate, otherwise we would have to calculate corner
    # coordinates outside the determined range, which could trigger certain edge cases
    latSpace = latSpaceCenter[:-1] + latStep/2
    lonSpace = lonSpaceCenter[:-1] + lonStep/2
    latSpaceCenter = latSpaceCenter[1:-1]
    lonSpaceCenter = lonSpaceCenter[1:-1]

    #latGrid, lonGrid = np.meshgrid(latSpace, lonSpace, indexing='ij') # 'indexing' not supported in np 1.6
    latGrid, lonGrid = np.dstack(np.meshgrid(latSpace, lonSpace)).T
    latGridCenter, lonGridCenter = np.dstack(np.meshgrid(latSpaceCenter, lonSpaceCenter)).T
    
    # do the actual resampling
    dataResampled = _resampleCenterData(latsCenter, lonsCenter, 
                                        data, latSpaceCenter, lonSpaceCenter, latStep, lonStep,
                                        method)

    # mask grid points which are outside the outline
    # This is needed as 'linear' and 'cubic' only mask points outside the *convex hull*,
    # which is not enough as we have concave forms. In those corner cases the data is interpolated.
    # With 'nearest', nothing is masked.
    # With 'mean', there is no inter/extrapolation, so we can skip the additional masking.
    if method != 'mean':
        # Based on the masked grid points the data is masked if any of its 4 corner points is masked.
        outlineLatLon = outlineLatLonFn()
        latLonGridFlat = np.asarray([np.ravel(latGrid), np.ravel(lonGrid)]).T
        isOutside = ~pointsInsidePolygon(latLonGridFlat, outlineLatLon).reshape(latGrid.shape)      
        mask = np.logical_or.reduce((isOutside[:-1,:-1], isOutside[1:,:-1], isOutside[:-1,1:], isOutside[1:,1:]))
        dataResampled[mask] = np.nan
        
    # rotate back coordinates if previously rotated
    if containsPole:
        angle = -90
        axis = [1,0,0]
        latGridFlat, lonGridFlat = rotatePole(np.deg2rad(latGrid.ravel()), np.deg2rad(lonGrid.ravel()), 
                                              altitude, angle=angle, axis=axis)
        latGrid = np.rad2deg(latGridFlat.reshape(latGrid.shape))
        lonGrid = np.rad2deg(lonGridFlat.reshape(latGrid.shape))
        
        latGridCenterFlat, lonGridCenterFlat = rotatePole(np.deg2rad(latGridCenter.ravel()), np.deg2rad(lonGridCenter.ravel()), 
                                                          altitude, angle=angle, axis=axis)
        latGridCenter = np.rad2deg(latGridCenterFlat.reshape(latGridCenter.shape))
        lonGridCenter = np.rad2deg(lonGridCenterFlat.reshape(latGridCenter.shape))
    elif containsDiscontinuity:
        angle = 180
        lonGrid = Angle((lonGrid + angle) * u.deg).wrap_at(angle * u.deg).degree
        lonGridCenter = Angle((lonGridCenter + angle) * u.deg).wrap_at(angle * u.deg).degree
        
    return latGrid, lonGrid, latGridCenter, lonGridCenter, dataResampled

def fixedGrid(pxPerDeg, latMin, latMax, lonMin, lonMax):
    """
    Aligns the given bounding box to a fixed plate carree grid as defined
    by `pxPerDeg`.
    
    :param lonMin,lonMax: must NOT contain the discontinuity
    """
    latPxPerDeg, lonPxPerDeg = pxPerDeg
    nLatAll = latPxPerDeg*180 + 1
    nLonAll = lonPxPerDeg*360 + 1
    latSpaceAll = np.linspace(-90, 90, int(round(nLatAll)))
    lonSpaceAll = np.linspace(-180, 180, int(round(nLonAll)))
    latMinInGrid = latSpaceAll[np.argmax(latSpaceAll > latMin) - 1]
    latMaxInGrid = latSpaceAll[np.argmax(latSpaceAll >= latMax)]
    lonMinInGrid = lonSpaceAll[np.argmax(lonSpaceAll > lonMin) - 1]
    lonMaxInGrid = lonSpaceAll[np.argmax(lonSpaceAll >= lonMax)]
    nLat = int(round(latPxPerDeg*(latMaxInGrid-latMinInGrid) + 1))
    nLon = int(round(lonPxPerDeg*(lonMaxInGrid-lonMinInGrid) + 1))
    return nLat, nLon, latMinInGrid, latMaxInGrid, lonMinInGrid, lonMaxInGrid

def _resampleCenterData(latsCenter, lonsCenter, centerData, latSpaceCenter, lonSpaceCenter, latStep, lonStep,
                        method):
    """
    :param method: binning: 'mean'; interpolation: 'nearest', 'linear', 'cubic';
                   Note that linear and cubic take considerably longer and use much more memory while
                   they don't bring any benefit over 'nearest' if the goal is downsampling.
    """
    if centerData.ndim == 2:
        scalarData = True
        centerData = centerData[...,None]
    else:
        scalarData = False
    
    # interpolate center data at grid point centers
    centerInLatsFlat, centerInLonsFlat = np.ravel(latsCenter), np.ravel(lonsCenter)
    centerNonNans = ~np.isnan(centerInLatsFlat)
    centerInLatsFlatFiltered = centerInLatsFlat[centerNonNans]
    centerInLonsFlatFiltered = centerInLonsFlat[centerNonNans]
    
    centerFlat = centerData.reshape(-1,centerData.shape[2])
    centerFlatFiltered = centerFlat[centerNonNans]
    
    if method in ['nearest', 'linear', 'cubic']:
        centerResampled = scipy.interpolate.griddata(
                              (centerInLatsFlatFiltered,centerInLonsFlatFiltered), centerFlatFiltered,
                              (latSpaceCenter[:,None], lonSpaceCenter[None,:]), method=method)
    
    elif method == 'mean':
        # this is about 20-50% slower than griddata's 'nearest'
        bins = (len(lonSpaceCenter), len(latSpaceCenter))
        # the bin egdes must be monotonically increasing, therefore we do that and flip it afterwards
        # so that latitudes are decreasing
        range_ = [[lonSpaceCenter[0]-lonStep/2, lonSpaceCenter[-1]+lonStep/2],
                  [latSpaceCenter[-1]+latStep/2, latSpaceCenter[0]-latStep/2]]    
        data = [centerFlatFiltered[:,d] for d in range(centerData.shape[2])]
                
        countAndData,_,_ = histogram2d(centerInLonsFlatFiltered, centerInLatsFlatFiltered, bins=bins, range=range_,
                                       weights=[None]+data)
        count = countAndData[0].T
        
        resampled = []
        for d in range(centerData.shape[2]):
            data = centerFlatFiltered[:,d]
            dataOut = countAndData[d+1].T
            dataOut[count==0.0] = np.nan
            with np.errstate(invalid='ignore'):
                dataOut /= count
            # flip so that latitudes are decreasing
            dataOut = np.flipud(dataOut)
            resampled.append(dataOut)
        centerResampled = np.dstack(resampled)
        
    elif method == 'median':
        raise NotImplementedError
        # https://stackoverflow.com/a/15488537
        # https://stackoverflow.com/a/10324083
        # http://fspaolo.net/code/bin-data.html       
        
    else:
        raise NotImplementedError
    
    assert centerResampled.shape == (len(latSpaceCenter), len(lonSpaceCenter), centerData.shape[2]),\
           str(centerResampled.shape) + ' != ' + str((len(latSpaceCenter), len(lonSpaceCenter), centerData.shape[2]))
    
    if scalarData:
        centerResampled = centerResampled.reshape(centerResampled.shape[0], centerResampled.shape[1])
        
    return centerResampled

def ResampleProvider(provider, **kw):
    """
    Wrap the given mapping provider by resampling every returned mapping.    
    
    :param provider: the provider to wrap
    
    See :func:`resample` for masking parameters.
    """
    resampleFn = partial(resample, **kw)
    class ResamplingProvider(object):          
        def get(self, *args, **kw):
            m = super(ResamplingProvider, self).get(*args, **kw)
            return resampleFn(m)
        
        def getById(self, *args, **kw):
            m = super(ResamplingProvider, self).getById(*args, **kw)
            return resampleFn(m)
            
        def getSequence(self, *args, **kw):
            m = super(ResamplingProvider, self).getSequence(*args, **kw)
            return map(resampleFn, m)
    
    provider = copy.copy(provider)  
    extend(provider, ResamplingProvider)
    return provider



# Copyright European Space Agency, 2013

from __future__ import absolute_import, print_function

import os
import time

import numpy as np
from numpy.testing.utils import assert_almost_equal
from astropy.wcs.wcs import WCS
from auromat.fits import readHeader
from auromat.coordinates.wcs import tan_pix2world
    
def wcsTest():
    wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
    header = readHeader(wcsPath)
    
    # the pixel coordinates of which we want the world coordinates
    x, y = np.meshgrid(np.arange(0,4000), np.arange(0,2000))
    x = x.ravel()
    y = y.ravel()
    
    # first, calculate using astropy as reference
    wcs = WCS(header)
    t0 = time.time()
    ra_astropy, dec_astropy = wcs.wcs_pix2world(x, y, 0)
    print('astropy wcs:', time.time()-t0, 's')
    
    # now, check against our own implementation
    #tan_pix2world(header, x, y, 0) # warmup
    t0 = time.time()
    ra, dec = tan_pix2world(header, x, y, 0)
    print('own wcs:', time.time()-t0, 's')
    
    assert_almost_equal([ra, dec], [ra_astropy, dec_astropy])

def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

if __name__ == '__main__':
    wcsTest()
    


#Copyright (C) 2011, the scikit-image team
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are
#met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 3. Neither the name of skimage nor the names of its contributors may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
#INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
#IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#POSSIBILITY OF SUCH DAMAGE.


import numpy as np
from numpy.lib.stride_tricks import as_strided

def view_as_blocks(arr_in, block_shape):
    """Block view of the input n-dimensional array (using re-striding).

    Blocks are non-overlapping views of the input array.
    
    This function is the same as skimage.util.shape.view_as_blocks except
    that it errors if the input error is not c-contiguous.
    
    :see: https://github.com/scikit-image/scikit-image/blob/master/skimage/util/shape.py
    :param ndarray arr_in: The n-dimensional input array (must be C-contiguous).
    :param tuple block_shape: The shape of the block. Each dimension must divide 
                              evenly into the corresponding dimensions of `arr_in`.
    :rtype: ndarray, block view of the input array.
    """
    # -- basic checks on arguments
    if not isinstance(block_shape, tuple):
        raise TypeError('block needs to be a tuple')

    block_shape = np.array(block_shape)
    if (block_shape <= 0).any():
        raise ValueError("'block_shape' elements must be strictly positive")

    if block_shape.size != arr_in.ndim:
        raise ValueError("'block_shape' must have the same length "
                         "as 'arr_in.shape'")

    arr_shape = np.array(arr_in.shape)
    if (arr_shape % block_shape).sum() != 0:
        raise ValueError("'block_shape' is not compatible with 'arr_in'")
    
    if not arr_in.flags['C_CONTIGUOUS']:
        raise ValueError("'arr_in' must be c-contiguous")

    # -- restride the array to build the block view
    new_shape = tuple(arr_shape / block_shape) + tuple(block_shape)
    new_strides = tuple(arr_in.strides * block_shape) + arr_in.strides

    arr_out = as_strided(arr_in, shape=new_shape, strides=new_strides)

    return arr_out


# Copyright European Space Agency, 2013

"""
Internal module used by the .draw module.
"""

from __future__ import division, absolute_import, print_function

from six import integer_types, string_types
import warnings
import datetime
from collections import namedtuple

import numpy as np
import numpy.ma as ma
from numpy.core.umath_tests import inner1d # vector-wise dot product

from matplotlib.lines import Line2D
from matplotlib.text import Text
from matplotlib.patches import Circle
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import brewer2mpl
from matplotlib.collections import PatchCollection
blue_red = brewer2mpl.get_map('RdBu', 'Diverging', 11, reverse=True).mpl_colormap

from auromat.util.image import image2mpl, loadImage
from auromat.mapping.mapping import BaseMapping, MappingCollection,\
    convertMappingToSM

class ColorMode:
    matplotlib = 'matplotlib'

def createPolygonsAndColors(latDeg, lonDeg, rgb, colorMode=None):
    """
    Returns polygons (in lat/lon coords) and a color for each polygon.
    
    :param latDeg: latitude for each pixel corner (h+1,w+1)
    :param lonDeg: longitude for each pixel corner (h+1,w+1)
    :param rgb: RGB array of (h,w,3) shape
    :param colorMode: 'matplotlib' normalizes colors to [0,1]
    :rtype: verts of shape (h*w,4,2), colors of shape (h*w,3)
    """
    latLonDeg = ma.dstack((latDeg,lonDeg))
    
    # adapted from matplotlib.collections.QuadMesh.convert_mesh_to_paths
    verts = ma.concatenate((
                latLonDeg[0:-1, 0:-1],
                latLonDeg[0:-1, 1:  ],
                latLonDeg[1:  , 1:  ],
                latLonDeg[1:  , 0:-1],
                #latLonDeg[0:-1, 0:-1] # matplotlib automatically closes the polygon
                ), axis=2)
    verts = verts.reshape(rgb.shape[0] * rgb.shape[1], 4, 2)
    
    if colorMode == ColorMode.matplotlib:
        rgb = image2mpl(rgb)
    
    colors = rgb.reshape(-1,3)
    
    return verts, colors
    
def filterNanPolygons(verts, colors):
    # when color is defined, then all its corners are also defined
    # according to BaseMapping guarantees, therefore we filter using color
    if ma.isMaskedArray(colors):
        hasNans = ma.getmaskarray(colors)[:,0]
    else:
        hasNans = np.isnan(colors[:,0])
    verts = verts[~hasNans]
    colors = colors[~hasNans]
    # we don't need masked arrays anymore after all invalid values are gone
    if ma.isMaskedArray(verts):
        verts = verts.data
    if ma.isMaskedArray(colors):
        colors = colors.data
    #assert np.all(~np.isnan(verts))
    #assert np.all(~np.isnan(colors))
    return verts, colors

def generatePolygonsFromMapping(mapping, colorMode=None, coordsFn=None):
    """
    :param colorMode: 'matplotlib' normalizes colors to [0,1]
    """
    coordsFn = _coordsFn(coordsFn)
    lats, lons = coordsFn(mapping)
    rgb = mapping.rgb
    verts, colors = createPolygonsAndColors(lats, lons, rgb, colorMode)
    verts, colors = filterNanPolygons(verts, colors)
    return verts, colors
    
def overlapPolygons(verts):
    """
    See https://github.com/matplotlib/matplotlib/issues/2823
    
    NOTE: the bigger the polygons, the higher the resulting error
    """
    # TODO overlapping leads to visible borders, esp. in resampled data
    # FIXME this will cause trouble in 180deg discontinuity area with fixDiscontinuityPolys
    a = 1.0 # the factor should be much lower but matplotlib seems to need it...
    verts[:,1,:] = verts[:,1,:] + a*(verts[:,1,:]-verts[:,0,:])
    verts[:,2,:] = verts[:,2,:] + a*(verts[:,2,:]-verts[:,3,:])
    verts[:,2,:] = verts[:,2,:] + a*(verts[:,2,:]-verts[:,1,:])
    verts[:,3,:] = verts[:,3,:] + a*(verts[:,3,:]-verts[:,0,:])
    return verts

def fixDiscontinuityPolys(verts, colors):
    """
    Splits polygons which cross the 180deg discontinuity into two parts.
    
    NOTE: The vertices must lie on a regular lat/lon grid!
    
    Currently not used, see comment on _drawPolygons.
     
    :param verts: (n,4,2) in degrees (:,:,0 = lat)
    :param colors: (n,3)
    """
    length = np.abs(verts[:,3,1] - verts[:,0,1])
    polysWithDiscont = verts[length>180]
    print(polysWithDiscont)
    return verts, colors

def _coordsFn(coordsFn):
    if coordsFn is None:
        coordsFn = lambda mapping: (mapping.lats, mapping.lons)
    return coordsFn

def _generatePolygonsFromMappingOrCollection(mappings, coordsFn=None):
    """
    replicates THEMIS drawing (higher elevation pixels have priority when mappings
    overlap) for MappingCollection's which have mayOverlap==True
        
    :param mappings: BaseMapping or MappingCollection
    :rtype: tuple (vertsArr, colorsArr)
    """
    coordsFn = _coordsFn(coordsFn)
        
    if isinstance(mappings, BaseMapping):
        mapping = mappings
        verts, colors = generatePolygonsFromMapping(mapping, ColorMode.matplotlib, coordsFn=coordsFn)
        vertsArr, colorsArr = [verts], [colors]
    
    elif isinstance(mappings, MappingCollection):
        vertsArr = []
        colorsArr = []
        elevationArr = []
        for mapping in mappings.mappings:
            if mappings.mayOverlap:
                # don't apply filtering yet, need to relate to elevation
                lats, lons = coordsFn(mapping)
                verts, colors = createPolygonsAndColors(lats, lons, mapping.rgb, ColorMode.matplotlib)
            else:
                verts, colors = generatePolygonsFromMapping(mapping, ColorMode.matplotlib, coordsFn=coordsFn)
            vertsArr.append(verts)
            colorsArr.append(colors)
            if mappings.mayOverlap:
                el = mapping.elevation.ravel()
                assert len(el) == len(verts) == len(colors)
                elevationArr.append(el)
        
        if mappings.mayOverlap and len(mappings)>0:
            # The polygons are sorted according to increasing elevation. If they
            # are drawn in the same order then the outermost parts of an ASI image will
            # be overdrawn by higher-elevation polygons of overlapping ASIs. The reasoning
            # is that low-elevation pixels are less important (due to the projection error)
            # than high-elevation pixels.
            verts = np.concatenate(vertsArr)
            colors = np.concatenate(colorsArr)
            elevation = np.concatenate(elevationArr)

            elSorted = np.argsort(elevation)
            verts, colors = filterNanPolygons(verts[elSorted], colors[elSorted])
            
            vertsArr = [verts]
            colorsArr = [colors]
    else:
        raise ValueError
    return vertsArr, colorsArr

def _addFigureBottomTitle(fig, title):
    fig.text(0.5, 0.02, title, horizontalalignment='center')
   
def _convertMappingsToSM(mappings):
    # for drawing, we pretend that Mlat/MLT are geographic coordinates
    # that way, we can reuse the regular drawing logic (esp. bounding box calculation etc.)    
    # convertMapping and convertMappingOrCollection map MLat/MLT to lat/lon coordinates   
    def convertMappingOrCollection(mapping):
        if isinstance(mapping, BaseMapping):
            newMapping = convertMappingToSM(mapping)
        elif isinstance(mapping, MappingCollection):
            mappings_ = list(map(convertMappingToSM, mapping.mappings))
            newMapping = MappingCollection(mappings_, mapping.identifier, mapping.mayOverlap)
        else:
            raise ValueError
        return newMapping
    
    if isinstance(mappings, list):
        newMappings = []
        for mapping in mappings:
            newMapping = convertMappingOrCollection(mapping)    
            newMappings.append(newMapping)
    else:
        newMappings = convertMappingOrCollection(mappings)
    
    return newMappings
    
def _formatMLT(smlon, format_='%H:%M:%S'):
    eps = 1e-10
    if smlon > 180:
        smlon = smlon-360
    mlt = smlon*(24/360) + 12
    if abs(mlt - 24) < eps:
        mlt = 0
    hours = int(mlt)
    minutes = int((mlt - hours)*60)
    seconds = int(((mlt - hours)*60 - minutes)*60)
    micros = int((((mlt - hours)*60 - minutes)*60 - seconds)*1e6)
    # round 02:59:59.999 up
    if micros > 999900:
        # this should be written more nicely..
        if seconds == 59:
            seconds = 0
            if minutes == 59:
                minutes = 0
                if hours == 23:
                    hours = 0
                else:
                    hours += 1
            else:
                minutes += 1
        else:
            seconds += 1
        micros = 0
    timestr = datetime.time(hours, minutes, seconds, micros).strftime(format_)
    return timestr

def _circles(x, y, s, ax, **kwargs):
    """
    Draw circles with given radius.

    :param x,y: coordinates of circles
    :type x,y: scalar or array_like, shape (n,)
    :param s: radius of circles
    :type s: scalar or array_like, shape (n,)
    :param Axes ax: axes to draw onto
    :param kwargs: additional keywords handed over to `~matplotlib.collections.Collection`
        eg. alpha, edgecolors, facecolors, linewidths, linestyles, norm, cmap
    :rtype: matplotlib.collections.PathCollection
    """
    if isinstance(x, integer_types) or isinstance(x, float):
        patches = [Circle((x, y), s),]
    elif isinstance(s, integer_types) or isinstance(s, float):
        patches = [Circle((x_,y_), s) for x_,y_ in zip(x,y)]
    else:
        patches = [Circle((x_,y_), s_) for x_,y_,s_ in zip(x,y,s)]
    collection = PatchCollection(patches, **kwargs)

    ax.add_collection(collection)
    return collection

def ensureContinuousPath(points):
    """
    Reorders a maximum of two logical segments in the given path whose points must
    be at most one (also diagonal) pixel apart.
    
    For example, the first segment goes from the middle to the right, and the
    second from the left to the middle. This function detects and swaps the two
    segments such that one continuous segment results.
    
    :param points: shape (n,2)
    :rtype: points with segments swapped, or original array if no swap necessary
    """
    vecs = points[1:] - points[:-1]
    lenSq = inner1d(vecs, vecs)
    jumps = lenSq > 2
    if np.any(jumps):
        jumpIdx = np.argmax(jumps)
        return np.concatenate((points[jumpIdx+1:], points[:jumpIdx+1]))
    else:
        return points

def _saveFig(outputFile, fig, widthPx=None, heightPx=None, dpi=None, format_=None, dontClose=False):
    assert not(widthPx and heightPx), 'Either specify widthPx OR heightPx, not both'
    assert not((widthPx or heightPx) and dpi), 'Either specify dpi OR widthPx OR heightPx'
    
    if widthPx:
        dpi = widthPx / fig.get_figwidth()
    elif heightPx:
        dpi = heightPx / fig.get_figheight()
    else:
        dpi = fig.get_dpi()
        
    fig.savefig(outputFile, dpi=dpi, facecolor=fig.get_facecolor(), format=format_)
        
    if not dontClose:
        plt.close(fig)

def loadFigImage(im):
    """
    Return a matplotlib `Figure` with the given raster image spanning the plot extents
    and with data coordinates equal to pixel coordinates. All axis and labels are
    hidden.
    
    :param im: either path to image file, or RGB image array in uint8 or uint16 type
    :rtype: tuple(Figure,Axes)
    """
    if isinstance(im, string_types):
        im = loadImage(im)
    im = image2mpl(im)
    h,w = im.shape[0], im.shape[1]
    dpi = 80
    fig = plt.figure(figsize=(w/dpi, h/dpi), dpi=dpi)
    ax = plt.Axes(fig, [0, 0, 1, 1])
    ax.set_xlim(0, w)
    ax.set_ylim(0, h)
    ax.invert_yaxis()
    ax.set_axis_off()
    fig.add_axes(ax)
    if im.ndim == 2:
        fig.figimage(im, cmap=cm.gray)
    else:
        fig.figimage(im)
    fig._dontSetColors = True

    return fig, ax

def _setMplColors(fig, ax, colors):
    if fig and hasattr(fig, '_dontSetColors'):
        return        
    if fig:
        fig.patch.set_facecolor(colors.facecolor)
    ax.set_axis_bgcolor(colors.facecolor)
    ax.patch.set_facecolor(colors.facecolor)
    
    ax.spines['bottom'].set_color(colors.textcolor)
    ax.spines['top'].set_color(colors.textcolor)
    ax.spines['left'].set_color(colors.textcolor)
    ax.spines['right'].set_color(colors.textcolor)
    for i in ax.get_xticklabels(): i.set_color(colors.textcolor)
    for i in ax.get_yticklabels(): i.set_color(colors.textcolor)
    for t in ax.xaxis.get_ticklines(): t.set_color(colors.textcolor)
    for t in ax.yaxis.get_ticklines(): t.set_color(colors.textcolor)
    for t in ax.xaxis.get_minorticklines(): t.set_color(colors.textcolor)
    for t in ax.yaxis.get_minorticklines(): t.set_color(colors.textcolor)
    ax.xaxis.label.set_color(colors.textcolor)
    ax.yaxis.label.set_color(colors.textcolor)
    for t in fig.texts: t.set_color(colors.textcolor)
    for t in ax.findobj(Text): t.set_color(colors.textcolor)
    
    if fig and hasattr(fig, '_applyMeridiansParallelsColors'):
        for t in ax.findobj(Line2D): t.set_color(colors.textcolor)
    
    # circles of polar maps
    if fig and hasattr(fig, '_applyRoundPolarMapColors'):
        for c in ax.findobj(Circle):
            c.set_edgecolor(colors.textcolor)
        
MplColors = namedtuple('MplColors', ['facecolor', 'textcolor'])

def _getMplColors(bgcolor='white', transparent=False):
    textcolor = 'white' if bgcolor == 'black' else 'black'
    facecolor = 'none' if transparent else bgcolor
    return MplColors(facecolor, textcolor)



# Copyright European Space Agency, 2013

"""
This module exports :class:`auromat.mapping.mapping.BaseMapping` objects
into NASA's CDF file format following the ISTP/IACG guidelines.
Each mapping is exported as a single self-contained file.

See http://cdf.gsfc.nasa.gov/html/CDF_docs.html and
http://spdf.gsfc.nasa.gov/sp_use_of_cdf.html for details.

Where no guidelines exist, the ones from the :mod:`auromat.export.netcdf`
module are used, in particular the `geospatial_*` attributes for describing
the bounding box, and the `crs` attributes for describing the coordinate
systems.
"""

from __future__ import division, absolute_import, print_function

import numpy as np
from spacepy import pycdf

from auromat.coordinates.transform import northGeomagneticPoleLocation
import sys

def write(outputPath, mapping, metadata={}, includeBounds=True, 
          includeMagCoords=True, includeGeoCoords=True, 
          compress=True, useTT2000=True):
    """
    
    :param str outputPath:
    :param auromat.mapping.mapping.BaseMapping mapping:
    :param dict metadata: additional metadata, overwrites `mapping.metadata` entries if existing,
                          a dictionary of root attributes, e.g.::
                          
                            {'Project': '..',
                             'Source_name': '..',
                             'Discipline': 'Space Physics>Magnetospheric Science',
                             'Descriptor': '..'
                            }
                          
                          See http://spdf.gsfc.nasa.gov/istp_guide/gattributes.html.
    :param bool includeBounds: stores the coordinates of each pixel corner (in addition to the center)
    :param bool includeMagCoords: include geomagnetic latitude-magnetic local time coordinates
    :param bool includeGeoCoords: include geodetic coordinates
    :param bool compress: use compression for variables
    :param bool useTT2000: 
        Whether to use the CDF_TIME_TT2000 type for storing times.
        If False, then the CDF_EPOCH type is used.
        See http://cdf.gsfc.nasa.gov/html/leapseconds.html for the advantages 
        of using CDF_TIME_TT2000.
        Note that this requires CDF 3.4.0 or higher both for reading and writing.
    """
    # for CDF_TIME_TT2000 backward compatible mode must be off
    pycdf.lib.set_backward(not useTT2000)
    
    if compress:
        compress = pycdf.const.GZIP_COMPRESSION
    else:
        compress = pycdf.const.NO_COMPRESSION
    
    with pycdf.CDF(outputPath, '') as root:            
        # ROOT ATTRIBUTES
        metadata = dict(list(mapping.metadata.items()) + list(metadata.items()))
        for k,v in metadata.items():
            if isinstance(v, bool):
                v = int(v)
            try:
                root.attrs[k] = v
            except TypeError:
                print('Cannot store global attribute "{}" with value {}'.format(k,repr(v)), file=sys.stderr) 
                raise
        
        # NODC conventions:
        root.attrs['geospatial_lat_min'] = mapping.boundingBox.latSouth
        root.attrs['geospatial_lat_max'] = mapping.boundingBox.latNorth
        root.attrs['geospatial_lon_min'] = mapping.boundingBox.lonWest
        root.attrs['geospatial_lon_max'] = mapping.boundingBox.lonEast
        root.attrs['geospatial_lat_units'] = 'degrees_north'
        root.attrs['geospatial_lon_units'] = 'degrees_east'
        
        # VARIABLES
        if useTT2000:
            root.new('Epoch', [mapping.photoTime], type=pycdf.const.CDF_TIME_TT2000)
        else:
            # uses CDF_EPOCH by default
            root['Epoch'] = [mapping.photoTime]
        time = root['Epoch']
        time.attrs['VAR_TYPE'] = 'support_data'
        
        if includeGeoCoords:
            root.new('lat', mapping.latsCenter[np.newaxis,:], compress=compress)
            lat = root['lat']
            lat.attrs['VAR_TYPE'] = 'data'
            lat.attrs['DEPEND_0'] = 'Epoch'
            lat.attrs['DEPEND_1'] = 'y_pixel'
            lat.attrs['DEPEND_2'] = 'x_pixel'
            lat.attrs['UNITS'] = 'degrees'
            lat.attrs['VALIDMIN'] = -90.0
            lat.attrs['VALIDMAX'] = 90.0
            lat.attrs['FIELDNAM'] = 'Latitude of pixel center'    
            lat.attrs['VAR_NOTES'] = 'Geodetic latitude'
            lat.attrs['crs'] = 'crs'
            
            root.new('lon', mapping.lonsCenter[np.newaxis,:], compress=compress)
            lon = root['lon']
            lon.attrs['VAR_TYPE'] = 'data'
            lon.attrs['DEPEND_0'] = 'Epoch'
            lon.attrs['DEPEND_1'] = 'y_pixel'
            lon.attrs['DEPEND_2'] = 'x_pixel'
            lon.attrs['UNITS'] = 'degrees'
            lon.attrs['VALIDMIN'] = -180.0
            lon.attrs['VALIDMAX'] = 180.0
            lon.attrs['FIELDNAM'] = 'Longitude of pixel center'    
            lon.attrs['VAR_NOTES'] = 'Geodetic longitude'
            lon.attrs['crs'] = 'crs'
            
            if includeBounds:
                lat.attrs['bounds'] = 'lat_bounds'
                lon.attrs['bounds'] = 'lon_bounds'
                
                root.new('lat_bounds', mapping.lats[np.newaxis,:], compress=compress)
                lat_bounds = root['lat_bounds']
                lat_bounds.attrs['VAR_TYPE'] = 'data'
                lat_bounds.attrs['DEPEND_0'] = 'Epoch'
                lat_bounds.attrs['DEPEND_1'] = 'y_corner'
                lat_bounds.attrs['DEPEND_2'] = 'x_corner'
                lat_bounds.attrs['UNITS'] = 'degrees'
                lat_bounds.attrs['VALIDMIN'] = -90.0
                lat_bounds.attrs['VALIDMAX'] = 90.0
                lat_bounds.attrs['FIELDNAM'] = 'Latitude of pixel corner'
                lat_bounds.attrs['VAR_NOTES'] = 'Geodetic latitude'
                lat_bounds.attrs['crs'] = 'crs'
                
                root.new('lon_bounds', mapping.lons[np.newaxis,:], compress=compress)
                lon_bounds = root['lon_bounds']
                lon_bounds.attrs['VAR_TYPE'] = 'data'
                lon_bounds.attrs['DEPEND_0'] = 'Epoch'
                lon_bounds.attrs['DEPEND_1'] = 'y_corner'
                lon_bounds.attrs['DEPEND_2'] = 'x_corner'
                lon_bounds.attrs['UNITS'] = 'degrees'
                lon_bounds.attrs['VALIDMIN'] = -180.0
                lon_bounds.attrs['VALIDMAX'] = 180.0
                lon_bounds.attrs['FIELDNAM'] = 'Longitude of pixel corner'    
                lon_bounds.attrs['VAR_NOTES'] = 'Geodetic longitude'
                lon_bounds.attrs['crs'] = 'crs'
        
        root.new('altitude', mapping.altitude * 1000, recVary=False)
        altitude = root['altitude']
        altitude.attrs['VAR_TYPE'] = 'support_data'
        altitude.attrs['UNITS'] = 'meters'
        altitude.attrs['FIELDNAM'] = 'Height above reference ellipsoid'
        altitude.attrs['crs'] = 'crs'
        
        if includeMagCoords:            
            mlats, mlts = mapping.mLatMltCenter
                        
            root.new('mlat', mlats[np.newaxis,:], compress=compress)
            mlat = root['mlat']
            mlat.attrs['VAR_TYPE'] = 'data'
            mlat.attrs['DEPEND_0'] = 'Epoch'
            mlat.attrs['DEPEND_1'] = 'y_pixel'
            mlat.attrs['DEPEND_2'] = 'x_pixel'
            mlat.attrs['UNITS'] = 'degrees'
            mlat.attrs['VALIDMIN'] = -90.0
            mlat.attrs['VALIDMAX'] = 90.0
            mlat.attrs['FIELDNAM'] = 'Geomagnetic latitude of pixel center'
            mlat.attrs['VAR_NOTES'] = ''
            mlat.attrs['crs'] = 'mcrs'
                        
            root.new('mlt', mlts[np.newaxis,:], compress=compress)
            mlt = root['mlt']
            mlt.attrs['VAR_TYPE'] = 'data'
            mlt.attrs['DEPEND_0'] = 'Epoch'
            mlt.attrs['DEPEND_1'] = 'y_center'
            mlt.attrs['DEPEND_2'] = 'x_center'
            mlt.attrs['UNITS'] = 'hours'
            mlt.attrs['VALIDMIN'] = 0.0
            mlt.attrs['VALIDMAX'] = 24.0
            mlt.attrs['FIELDNAM'] = 'Magnetic local time of pixel center'
            mlt.attrs['crs'] = 'mcrs'
            
            if includeBounds:
                mlat.attrs['bounds'] = 'mlat_bounds'
                mlt.attrs['bounds'] = 'mlt_bounds'
                
                mlats, mlts = mapping.mLatMlt
                
                root.new('mlat_bounds', mlats[np.newaxis,:], compress=compress)
                mlat_bounds = root['mlat_bounds']
                mlat_bounds.attrs['VAR_TYPE'] = 'data'
                mlat_bounds.attrs['DEPEND_0'] = 'Epoch'
                mlat_bounds.attrs['DEPEND_1'] = 'y_corner'
                mlat_bounds.attrs['DEPEND_2'] = 'x_corner'
                mlat_bounds.attrs['UNITS'] = 'degrees'
                mlat_bounds.attrs['VALIDMIN'] = -90.0
                mlat_bounds.attrs['VALIDMAX'] = 90.0
                mlat_bounds.attrs['FIELDNAM'] = 'Geomagnetic latitude of pixel corner'
                mlat_bounds.attrs['VAR_NOTES'] = ''
                mlat_bounds.attrs['crs'] = 'mcrs'
                            
                root.new('mlt_bounds', mlts[np.newaxis,:], compress=compress)
                mlt_bounds = root['mlt_bounds']
                mlt_bounds.attrs['VAR_TYPE'] = 'data'
                mlt_bounds.attrs['DEPEND_0'] = 'Epoch'
                mlt_bounds.attrs['DEPEND_1'] = 'y_corner'
                mlt_bounds.attrs['DEPEND_2'] = 'x_corner'
                mlt_bounds.attrs['UNITS'] = 'hours'
                mlt_bounds.attrs['VALIDMIN'] = 0.0
                mlt_bounds.attrs['VALIDMAX'] = 24.0
                mlt_bounds.attrs['FIELDNAM'] = 'Magnetic local time of pixel corner'
                mlt_bounds.attrs['crs'] = 'mcrs'
            
            magPoleLat, magPoleLon = northGeomagneticPoleLocation(mapping.photoTime)
            
            root.new('mcrs', 0, recVary=False) # holds no actual data
            mcrs = root['mcrs']
            mcrs.attrs['VAR_TYPE'] = 'support_data'
            mcrs.attrs['north_geomagnetic_pole_lat'] = magPoleLat
            mcrs.attrs['north_geomagnetic_pole_lon'] = magPoleLon
            mcrs.attrs['VAR_NOTES'] = 'Geocentric MLat/MLT system based on the given geomagnetic pole position'

        if np.any(mapping.img.mask):
            # CDF supports much more types than netCDF
            imgDtypeMap = {np.dtype(np.uint8): np.int16,  # no overhead (compared to a separate int8 mask)
                           np.dtype(np.uint16): np.int32, # 1 Byte overhead per pixel  
                           np.dtype(np.uint32): np.int64  # 2 Byte overhead per pixel
                           }
            if not mapping.img.dtype in imgDtypeMap.keys():
                raise NotImplementedError('Image data type not supported: ' + str(mapping.img.dtype))
            
            imgDtype = imgDtypeMap[mapping.img.dtype]
            imgFillval = imgDtype(np.iinfo(imgDtype).min)
            img_ = mapping.img.astype(imgDtype).filled(imgFillval)
        else:
            imgDtype = mapping.img.dtype
            img_ = img_.data
            imgFillval = None

        if img_.shape[2] == 1:
            bands = ['img']
        elif img_.shape[2] == 3:
            bands = ['img_red', 'img_green', 'img_blue']
        else:
            raise NotImplementedError

        for i, band in enumerate(bands):
            root.new(band, img_[np.newaxis,:,:,i], compress=compress)
            img = root[band]
            img.attrs['VAR_TYPE'] = 'data'
            img.attrs['DEPEND_0'] = 'Epoch'
            img.attrs['DEPEND_1'] = 'y_pixel'
            img.attrs['DEPEND_2'] = 'x_pixel'        
            img.attrs['FIELDNAM'] = ''
            img.attrs['VALIDMIN'] = np.iinfo(mapping.img.dtype).min
            img.attrs['VALIDMAX'] = np.iinfo(mapping.img.dtype).max
            if imgFillval:
                img.attrs['FILLVAL'] = imgFillval
            img.attrs['UNITS'] = 'unitless'
                    
        zen = 90 - mapping.elevation[np.newaxis,:].astype(np.float32)
        root.new('zenith_angle', zen, compress=compress)
        zenith_angle = root['zenith_angle']
        zenith_angle.attrs['VAR_TYPE'] = 'data'
        zenith_angle.attrs['DEPEND_0'] = 'Epoch'
        zenith_angle.attrs['DEPEND_1'] = 'y_pixel'
        zenith_angle.attrs['DEPEND_2'] = 'x_pixel'
        zenith_angle.attrs['UNITS'] = 'degrees'
        zenith_angle.attrs['VALIDMIN'] = 0.0
        zenith_angle.attrs['VALIDMAX'] = 90.0
        zenith_angle.attrs['FIELDNAM'] = 'Absolute sensor zenith angle of pixel center'

        root['camera_pos'] = np.array([mapping.cameraPosGCRS])
        cameraPos = root['camera_pos']
        cameraPos.attrs['VAR_TYPE'] = 'support_data'
        cameraPos.attrs['DEPEND_0'] = 'Epoch'
        cameraPos.attrs['UNITS'] = 'kilometers'
        cameraPos.attrs['FIELDNAM'] = 'Camera position in cartesian GCRS coordinates'
        cameraPos.attrs['VAR_NOTES'] = 'Axis order: xyz'

        root.new('crs', 0, recVary=False) # holds no actual data
        crs = root['crs']
        crs.attrs['VAR_TYPE'] = 'support_data'
        crs.attrs['semi_major_axis'] = 6378137.0
        crs.attrs['inverse_flattening'] = 298.257223563
        crs.attrs['VAR_NOTES'] = 'Geographic Coordinate System, WGS 84'



# Copyright European Space Agency, 2013

from __future__ import absolute_import, print_function

from nose.plugins.attrib import attr
import os
from time import sleep
from auromat.mapping.spacecraft import getMappingSequence

@attr('slow')
def parallelTest():
    n = 5
    imagePaths = [getResourcePath('ISS030-E-102170_dc.jpg')]*n
    wcsPaths = [getResourcePath('ISS030-E-102170_dc.wcs')]*n
    for mapping in getMappingSequence(imagePaths, wcsPaths, parallel=True):
        # simulate heavy work
        print(mapping.arcSecPerPx)
        sleep(10)
        
    
def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

if __name__ == '__main__':
    parallelTest()


# Copyright European Space Agency, 2013

"""
A little helper module to create movies out of images using ffmpeg.
"""

from __future__ import absolute_import, print_function

import os
import errno
import tempfile
import shutil
import subprocess

def createMovie(moviePath, imagePaths, frameRate=25, crf=18, maxBitrate=2000, width=None, height=None,
                tempFolder=None):
    """
    
    :param moviePath: path with .mp4 or .webm ending, will be overridden if existing
    :param imagePaths: paths of images in correct order
    :param frameRate: the frame rate of the movie
    :param crf: -crf parameter of ffmpeg, lower is better
    :param maxBitrate: maximum bitrate in kbit/s
    :param width, height: width and height of movie in pixels; if not given, then image dimensions are used
    :param tempFolder: the folder to store temporary data in (symlinks), will be removed afterwards
    """
    if not tempFolder:
        tempFolder = tempfile.mkdtemp()
    elif not os.path.exists(tempFolder):
        os.makedirs(tempFolder)
    
    ext = os.path.splitext(imagePaths[0])[1]
    framePaths = [os.path.join(tempFolder, str(frame) + ext) for frame in range(len(imagePaths))]
    
    movieExt = os.path.splitext(moviePath)[1]
    
    try:
        # create symlinks in the temporary folder to the input images
        for imagePath, framePath in zip(imagePaths, framePaths):     
            # TODO doesn't work on Windows       
            os.symlink(imagePath, framePath)
        
        # run ffmpeg
        args = ['ffmpeg', '-y', '-r', str(frameRate), '-i', os.path.join(tempFolder, '%d' + ext),
                '-r', str(frameRate)]
        
        if width or height:
            if width and height:
                s = '{}:{}'.format(width, height)
            elif width:
                # makes sure that height is divisible by 2
                s = '{}:trunc(ow/a/2)*2'.format(width)
            elif height:
                # makes sure that width is divisible by 2
                s = 'trunc(oh/a/2)*2:{}'.format(height)
            args += ['-vf', 'scale=' + s]
        
        if movieExt == '.mp4':
            args += ['-codec:v', 'libx264', '-preset', 'slow', '-profile:v', 'baseline',
                     '-pix_fmt', 'yuv420p', '-movflags', 'faststart'] # faststart needs ffmpeg >= 1.0
            
        elif movieExt == '.webm':
            args += ['-codec:v', 'libvpx', '-quality', 'good', '-cpu-used', '0',
                     '-b:v', str(maxBitrate) + 'k', '-qmin', '10', '-qmax', '42']
        else:
            raise NotImplementedError('Unsupported video format: ' + ext)
        
        args += ['-crf', str(crf), '-maxrate', str(maxBitrate) + 'k', 
                 '-bufsize', str(maxBitrate*2) + 'k']
        args += [moviePath]
        
        print('running ' + ' '.join(args))
                
        try:
            process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise RuntimeError('ffmpeg could not be launched. Make sure it is in the PATH!')
            else:
                raise
        
        stdout, _ = process.communicate()
            
        if process.returncode != 0:
            raise RuntimeError('ffmpeg returned exit code ' + str(process.returncode) + '; cmd line: ' + ' '.join(args) + '; ' +
                               'stdout&stderr output follows: ' + stdout.decode('ascii'))
    finally:
        shutil.rmtree(tempFolder)
    


# Copyright European Space Agency, 2013

from __future__ import division

import unittest
import numpy as np
from numpy.testing.utils import assert_array_almost_equal, assert_array_equal
import datetime

from auromat.coordinates.transform import geodetic2EcefZero,\
    ecef2Geodetic, geo_to_sm, j2000_to_geo, j2000_to_sm, gei_to_gse, gse_to_gsm, gsm_to_sm,\
    gei_to_geo, geo_to_gei, cartesian_to_spherical, spherical_to_cartesian,\
    _cartesian_to_spherical_np, _cartesian_to_spherical_ne,\
    _spherical_to_cartesian_np, _spherical_to_cartesian_ne

class Test(unittest.TestCase):

    def testCartesianSpherical(self):
        h, w = 20, 10
        x = np.random.rand(h, w)
        y = np.random.rand(h, w)
        z = np.random.rand(h, w)
        r, lat, lon = cartesian_to_spherical(x, y, z)
        xr, yr, zr = spherical_to_cartesian(r, lat, lon)
        assert_array_almost_equal(xr, x)
        assert_array_almost_equal(yr, y)
        assert_array_almost_equal(zr, z)
        
        xyz = spherical_to_cartesian(r, lat, lon, astuple=False)
        assert_array_almost_equal(xyz, np.dstack((x,y,z)))
        
    def testCartesianSphericalShape(self):
        h, w = 20, 10
        x = np.random.rand(h, w)
        y = np.random.rand(h, w)
        z = np.random.rand(h, w)
        
        r, lat, lon = _cartesian_to_spherical_np(x, y, z)
        assert_array_equal(r.shape, x.shape)
        assert_array_equal(lat.shape, r.shape)
        assert_array_equal(lon.shape, r.shape)
        
        r, lat, lon = _cartesian_to_spherical_ne(x, y, z)
        assert_array_equal(r.shape, x.shape)
        assert_array_equal(lat.shape, r.shape)
        assert_array_equal(lon.shape, r.shape)
        
        xr, yr, zr = _spherical_to_cartesian_np(r, lat, lon)
        assert_array_equal(xr.shape, r.shape)
        assert_array_equal(yr.shape, xr.shape)
        assert_array_equal(zr.shape, xr.shape)
        
        xr, yr, zr = _spherical_to_cartesian_ne(r, lat, lon)
        assert_array_equal(xr.shape, r.shape)
        assert_array_equal(yr.shape, xr.shape)
        assert_array_equal(zr.shape, xr.shape)
        
        xyz = _spherical_to_cartesian_np(r, lat, lon, astuple=False)
        assert_array_equal(xyz.shape, r.shape + (3,))
        
        xyz = _spherical_to_cartesian_ne(r, lat, lon, astuple=False)
        assert_array_equal(xyz.shape, r.shape + (3,))
        
        xyz = _spherical_to_cartesian_np(r.ravel(), lat.ravel(), lon.ravel(), astuple=False)
        assert_array_equal(xyz.shape, r.ravel().shape + (3,))
        
        xyz = _spherical_to_cartesian_ne(r.ravel(), lat.ravel(), lon.ravel(), astuple=False)
        assert_array_equal(xyz.shape, r.ravel().shape + (3,))

    def testGeodeticConversions(self):
        for lat in np.linspace(-89.9,89.9):
            for lon in np.linspace(-179.9,179.9):
                x, y, z = geodetic2EcefZero(np.deg2rad(lat), np.deg2rad(lon))
                r = np.rad2deg(ecef2Geodetic(x, y, z))
                assert_array_almost_equal(r,[lat,lon], 11)
        
    def testGeodeticConversionsArray(self):
        lat, lon = np.mgrid[-89:89:5,-179:179:5]
        x, y, z = geodetic2EcefZero(np.deg2rad(lat), np.deg2rad(lon))
    
        r = ecef2Geodetic(x, y, z)
        #print np.rad2deg(r)
        assert_array_almost_equal(np.rad2deg(r),[lat,lon], 11)
    
    # the following values were calculated using SSCWeb
    # with input: lat=50, lon=-100, r=1, date as below
    # http://sscweb.gsfc.nasa.gov/cgi-bin/CoordCalculator.cgi
    date = datetime.datetime(2012,1,25,9,26,55)
    geo = [[-0.11, -0.63, 0.77]]
    j2000 = [[-0.62, 0.16, 0.77]]
    gei = [[-0.62, 0.16, 0.77]]
    gse = [[-0.72, -0.26, 0.64]]
    gsm = [[-0.72, -0.30, 0.62]]
    sm = [[-0.43, -0.30, 0.85]]
    
    # the following 4 tests use each a single rotation matrix
    def testGeiToGeo(self):
        geo = gei_to_geo(self.date, self.gei)
        assert_array_almost_equal(geo, self.geo, 2)
        
    def testGeiToGse(self):
        gse = gei_to_gse(self.date, self.gei)
        assert_array_almost_equal(gse, self.gse, 2)
        
    def testGseToGsm(self):
        gsm = gse_to_gsm(self.date, self.gse)
        assert_array_almost_equal(gsm, self.gsm, 2)
        
    def testGsmToSm(self):
        sm = gsm_to_sm(self.date, self.gsm)
        assert_array_almost_equal(sm, self.sm, 2)
        
    # the following test uses a reverse transformation
    def testGeoToGei(self):
        gei = geo_to_gei(self.date, self.geo)
        assert_array_almost_equal(gei, self.gei, 2)

    # the following 3 tests use several multiplied rotation matrices
    def testJ2000ToGeo(self):
        geo = j2000_to_geo(self.date, self.j2000)
        assert_array_almost_equal(geo, self.geo, 2)
        
    def testJ2000ToSm(self):
        sm = j2000_to_sm(self.date, self.j2000)
        assert_array_almost_equal(sm, self.sm, 2)
    
    def testGeoToSm(self):
        sm = geo_to_sm(self.date, self.geo)
        assert_array_almost_equal(sm, self.sm, 2)
